@@ -6,14 +6,19 @@
 
 #include "vm/DateTime.h"
 
-#include "mozilla/ArrayUtils.h"
+#if JS_HAS_INTL_API
+#  include "mozilla/intl/ICU4CGlue.h"
+#  include "mozilla/intl/TimeZone.h"
+#endif
 #include "mozilla/ScopeExit.h"
+#include "mozilla/Span.h"
 #include "mozilla/TextUtils.h"
-#include "mozilla/Unused.h"
 
 #include <algorithm>
 #include <cstdlib>
 #include <cstring>
+#include <iterator>
+#include <string_view>
 #include <time.h>
 
 #if !defined(XP_WIN)
@@ -21,29 +26,38 @@
 #  include <unistd.h>
 #endif /* !defined(XP_WIN) */
 
-#include "jsutil.h"
-
+#if JS_HAS_INTL_API
+#  include "builtin/intl/FormatBuffer.h"
+#endif
+#include "js/AllocPolicy.h"
 #include "js/Date.h"
+#include "js/GCAPI.h"
+#include "js/Utility.h"
+#include "js/Vector.h"
 #include "threading/ExclusiveData.h"
 
-#if ENABLE_INTL_API && !MOZ_SYSTEM_ICU
-#  include "unicode/basictz.h"
-#  include "unicode/locid.h"
-#endif /* ENABLE_INTL_API && !MOZ_SYSTEM_ICU */
-
-#if ENABLE_INTL_API && (!MOZ_SYSTEM_ICU || defined(ICU_TZ_HAS_RECREATE_DEFAULT))
-#  include "unicode/timezone.h"
-#  include "unicode/unistr.h"
-#endif /* ENABLE_INTL_API && (!MOZ_SYSTEM_ICU || \
-          defined(ICU_TZ_HAS_RECREATE_DEFAULT)) */
-
 #include "util/Text.h"
 #include "vm/MutexIDs.h"
+#include "vm/Realm.h"
+
+/* static */
+js::DateTimeInfo::ForceUTC js::DateTimeInfo::forceUTC(JS::Realm* realm) {
+  return realm->creationOptions().forceUTC() ? DateTimeInfo::ForceUTC::Yes
+                                             : DateTimeInfo::ForceUTC::No;
+}
 
 static bool ComputeLocalTime(time_t local, struct tm* ptm) {
+  // Neither localtime_s nor localtime_r are required to act as if tzset has
+  // been called, therefore we need to explicitly call it to ensure any time
+  // zone changes are correctly picked up.
+
 #if defined(_WIN32)
+  _tzset();
   return localtime_s(ptm, &local) == 0;
 #elif defined(HAVE_LOCALTIME_R)
+#  ifndef __wasi__
+  tzset();
+#  endif
   return localtime_r(&local, ptm);
 #else
   struct tm* otm = localtime(&local);
@@ -157,25 +171,45 @@ static int32_t UTCToLocalStandardOffsetSeconds() {
   return local_secs - (utc_secs + SecondsPerDay);
 }
 
-bool js::DateTimeInfo::internalUpdateTimeZoneAdjustment(
-    ResetTimeZoneMode mode) {
+void js::DateTimeInfo::internalResetTimeZone(ResetTimeZoneMode mode) {
+  // Nothing to do when an update request is already enqueued.
+  if (timeZoneStatus_ == TimeZoneStatus::NeedsUpdate) {
+    return;
+  }
+
+  // Mark the state as needing an update, but defer the actual update until it's
+  // actually needed to delay any system calls to the last possible moment. This
+  // is beneficial when this method is called during start-up, because it avoids
+  // main-thread I/O blocking the process.
+  if (mode == ResetTimeZoneMode::ResetEvenIfOffsetUnchanged) {
+    timeZoneStatus_ = TimeZoneStatus::NeedsUpdate;
+  } else {
+    timeZoneStatus_ = TimeZoneStatus::UpdateIfChanged;
+  }
+}
+
+void js::DateTimeInfo::updateTimeZone() {
+  MOZ_ASSERT(timeZoneStatus_ != TimeZoneStatus::Valid);
+
+  bool updateIfChanged = timeZoneStatus_ == TimeZoneStatus::UpdateIfChanged;
+
+  timeZoneStatus_ = TimeZoneStatus::Valid;
+
   /*
    * The difference between local standard time and UTC will never change for
    * a given time zone.
    */
-  utcToLocalStandardOffsetSeconds_ = UTCToLocalStandardOffsetSeconds();
+  int32_t newOffset = UTCToLocalStandardOffsetSeconds();
 
-  int32_t newTZA = utcToLocalStandardOffsetSeconds_ * msPerSecond;
-  if (mode == ResetTimeZoneMode::DontResetIfOffsetUnchanged &&
-      newTZA == localTZA_) {
-    return false;
+  if (updateIfChanged && newOffset == utcToLocalStandardOffsetSeconds_) {
+    return;
   }
 
-  localTZA_ = newTZA;
+  utcToLocalStandardOffsetSeconds_ = newOffset;
 
   dstRange_.reset();
 
-#if ENABLE_INTL_API && !MOZ_SYSTEM_ICU
+#if JS_HAS_INTL_API
   utcRange_.reset();
   localRange_.reset();
 
@@ -187,27 +221,44 @@ bool js::DateTimeInfo::internalUpdateTimeZoneAdjustment(
     timeZone_ = nullptr;
   }
 
+  timeZoneId_ = nullptr;
   standardName_ = nullptr;
   daylightSavingsName_ = nullptr;
-#endif /* ENABLE_INTL_API && !MOZ_SYSTEM_ICU */
+#endif /* JS_HAS_INTL_API */
 
-  return true;
+  // Propagate the time zone change to ICU, too.
+  {
+    // Tell the analysis calling into ICU cannot GC.
+    JS::AutoSuppressGCAnalysis nogc;
+
+    internalResyncICUDefaultTimeZone();
+  }
 }
 
-js::DateTimeInfo::DateTimeInfo() {
-  internalUpdateTimeZoneAdjustment(
-      ResetTimeZoneMode::ResetEvenIfOffsetUnchaged);
+js::DateTimeInfo::DateTimeInfo(bool forceUTC) : forceUTC_(forceUTC) {
+  // Set the time zone status into the invalid state, so we compute the actual
+  // defaults on first access. We don't yet want to initialize neither <ctime>
+  // nor ICU's time zone classes, because that may cause I/O operations slowing
+  // down the JS engine initialization, which we're currently in the middle of.
+  timeZoneStatus_ = TimeZoneStatus::NeedsUpdate;
 }
 
 js::DateTimeInfo::~DateTimeInfo() = default;
 
 int64_t js::DateTimeInfo::toClampedSeconds(int64_t milliseconds) {
   int64_t seconds = milliseconds / msPerSecond;
+  int64_t millis = milliseconds % msPerSecond;
+
+  // Round towards the start of time.
+  if (millis < 0) {
+    seconds -= 1;
+  }
+
   if (seconds > MaxTimeT) {
     seconds = MaxTimeT;
   } else if (seconds < MinTimeT) {
     /* Go ahead a day to make localtime work (does not work with 0). */
-    seconds = SecondsPerDay;
+    seconds = MinTimeT + SecondsPerDay;
   }
   return seconds;
 }
@@ -216,18 +267,10 @@ int32_t js::DateTimeInfo::computeDSTOffsetMilliseconds(int64_t utcSeconds) {
   MOZ_ASSERT(utcSeconds >= MinTimeT);
   MOZ_ASSERT(utcSeconds <= MaxTimeT);
 
-#if ENABLE_INTL_API && !MOZ_SYSTEM_ICU
-  UDate date = UDate(utcSeconds * msPerSecond);
-  constexpr bool dateIsLocalTime = false;
-  int32_t rawOffset, dstOffset;
-  UErrorCode status = U_ZERO_ERROR;
+#if JS_HAS_INTL_API
+  int64_t utcMilliseconds = utcSeconds * msPerSecond;
 
-  timeZone()->getOffset(date, dateIsLocalTime, rawOffset, dstOffset, status);
-  if (U_FAILURE(status)) {
-    return 0;
-  }
-
-  return dstOffset;
+  return timeZone()->GetDSTOffsetMs(utcMilliseconds).unwrapOr(0);
 #else
   struct tm tm;
   if (!ComputeLocalTime(static_cast<time_t>(utcSeconds), &tm)) {
@@ -249,8 +292,8 @@ int32_t js::DateTimeInfo::computeDSTOffsetMilliseconds(int64_t utcSeconds) {
     diff -= SecondsPerDay;
   }
 
-  return diff * msPerSecond;
-#endif /* ENABLE_INTL_API && !MOZ_SYSTEM_ICU */
+  return diff * int32_t(msPerSecond);
+#endif /* JS_HAS_INTL_API */
 }
 
 int32_t js::DateTimeInfo::internalGetDSTOffsetMilliseconds(
@@ -286,7 +329,7 @@ int32_t js::DateTimeInfo::getOrComputeValue(RangeCache& range, int64_t seconds,
 
   if (range.startSeconds <= seconds) {
     int64_t newEndSeconds =
-        Min(range.endSeconds + RangeExpansionAmount, MaxTimeT);
+        std::min({range.endSeconds + RangeExpansionAmount, MaxTimeT});
     if (newEndSeconds >= seconds) {
       int32_t endOffsetMilliseconds = (this->*compute)(newEndSeconds);
       if (endOffsetMilliseconds == range.offsetMilliseconds) {
@@ -310,7 +353,7 @@ int32_t js::DateTimeInfo::getOrComputeValue(RangeCache& range, int64_t seconds,
   }
 
   int64_t newStartSeconds =
-      Max<int64_t>(range.startSeconds - RangeExpansionAmount, MinTimeT);
+      std::max<int64_t>({range.startSeconds - RangeExpansionAmount, MinTimeT});
   if (newStartSeconds <= seconds) {
     int32_t startOffsetMilliseconds = (this->*compute)(newStartSeconds);
     if (startOffsetMilliseconds == range.offsetMilliseconds) {
@@ -358,55 +401,23 @@ void js::DateTimeInfo::RangeCache::sanityCheck() {
   assertRange(oldStartSeconds, oldEndSeconds);
 }
 
-#if ENABLE_INTL_API && !MOZ_SYSTEM_ICU
+#if JS_HAS_INTL_API
 int32_t js::DateTimeInfo::computeUTCOffsetMilliseconds(int64_t localSeconds) {
   MOZ_ASSERT(localSeconds >= MinTimeT);
   MOZ_ASSERT(localSeconds <= MaxTimeT);
 
-  UDate date = UDate(localSeconds * msPerSecond);
+  int64_t localMilliseconds = localSeconds * msPerSecond;
 
-  // ES2019 draft rev 0ceb728a1adbffe42b26972a6541fd7f398b1557
-  //
-  // 20.3.1.7 LocalTZA
-  //
-  // If |localSeconds| represents either a skipped (at a positive time zone
-  // transition) or repeated (at a negative time zone transition) locale
-  // time, it must be interpreted as a time value before the transition.
-  constexpr int32_t skippedTime = icu::BasicTimeZone::kFormer;
-  constexpr int32_t repeatedTime = icu::BasicTimeZone::kFormer;
-
-  int32_t rawOffset, dstOffset;
-  UErrorCode status = U_ZERO_ERROR;
-
-  // All ICU TimeZone classes derive from BasicTimeZone, so we can safely
-  // perform the static_cast.
-  // Once <https://unicode-org.atlassian.net/browse/ICU-13705> is fixed we
-  // can remove this extra cast.
-  auto* basicTz = static_cast<icu::BasicTimeZone*>(timeZone());
-  basicTz->getOffsetFromLocal(date, skippedTime, repeatedTime, rawOffset,
-                              dstOffset, status);
-  if (U_FAILURE(status)) {
-    return 0;
-  }
-
-  return rawOffset + dstOffset;
+  return timeZone()->GetUTCOffsetMs(localMilliseconds).unwrapOr(0);
 }
 
 int32_t js::DateTimeInfo::computeLocalOffsetMilliseconds(int64_t utcSeconds) {
   MOZ_ASSERT(utcSeconds >= MinTimeT);
   MOZ_ASSERT(utcSeconds <= MaxTimeT);
 
-  UDate date = UDate(utcSeconds * msPerSecond);
-  constexpr bool dateIsLocalTime = false;
-  int32_t rawOffset, dstOffset;
-  UErrorCode status = U_ZERO_ERROR;
+  UDate utcMilliseconds = UDate(utcSeconds * msPerSecond);
 
-  timeZone()->getOffset(date, dateIsLocalTime, rawOffset, dstOffset, status);
-  if (U_FAILURE(status)) {
-    return 0;
-  }
-
-  return rawOffset + dstOffset;
+  return timeZone()->GetOffsetMs(utcMilliseconds).unwrapOr(0);
 }
 
 int32_t js::DateTimeInfo::internalGetOffsetMilliseconds(int64_t milliseconds,
@@ -419,11 +430,9 @@ int32_t js::DateTimeInfo::internalGetOffsetMilliseconds(int64_t milliseconds,
                                  &DateTimeInfo::computeUTCOffsetMilliseconds);
 }
 
-bool js::DateTimeInfo::internalTimeZoneDisplayName(char16_t* buf, size_t buflen,
-                                                   int64_t utcMilliseconds,
-                                                   const char* locale) {
-  MOZ_ASSERT(buf != nullptr);
-  MOZ_ASSERT(buflen > 0);
+bool js::DateTimeInfo::internalTimeZoneDisplayName(
+    TimeZoneDisplayNameVector& result, int64_t utcMilliseconds,
+    const char* locale) {
   MOZ_ASSERT(locale != nullptr);
 
   // Clear any previously cached names when the default locale changed.
@@ -437,117 +446,123 @@ bool js::DateTimeInfo::internalTimeZoneDisplayName(char16_t* buf, size_t buflen,
     daylightSavingsName_.reset();
   }
 
-  bool daylightSavings = internalGetDSTOffsetMilliseconds(utcMilliseconds) != 0;
+  using DaylightSavings = mozilla::intl::TimeZone::DaylightSavings;
 
-  JS::UniqueTwoByteChars& cachedName =
-      daylightSavings ? daylightSavingsName_ : standardName_;
+  auto daylightSavings = internalGetDSTOffsetMilliseconds(utcMilliseconds) != 0
+                             ? DaylightSavings::Yes
+                             : DaylightSavings::No;
+
+  JS::UniqueTwoByteChars& cachedName = (daylightSavings == DaylightSavings::Yes)
+                                           ? daylightSavingsName_
+                                           : standardName_;
   if (!cachedName) {
     // Retrieve the display name for the given locale.
-    icu::UnicodeString displayName;
-    timeZone()->getDisplayName(daylightSavings, icu::TimeZone::LONG,
-                               icu::Locale(locale), displayName);
 
-    size_t capacity = displayName.length() + 1;  // Null-terminate.
-    JS::UniqueTwoByteChars displayNameChars(js_pod_malloc<char16_t>(capacity));
-    if (!displayNameChars) {
+    intl::FormatBuffer<char16_t, 0, js::SystemAllocPolicy> buffer;
+    if (timeZone()->GetDisplayName(locale, daylightSavings, buffer).isErr()) {
       return false;
     }
 
-    // Copy the display name. This operation always succeeds because the
-    // destination buffer is large enough to hold the complete string.
-    UErrorCode status = U_ZERO_ERROR;
-    displayName.extract(displayNameChars.get(), capacity, status);
-    MOZ_ASSERT(U_SUCCESS(status));
-    MOZ_ASSERT(displayNameChars[capacity - 1] == '\0');
-
-    cachedName = std::move(displayNameChars);
+    cachedName = buffer.extractStringZ();
+    if (!cachedName) {
+      return false;
+    }
   }
-
-  // Return an empty string if the display name doesn't fit into the buffer.
-  size_t length = js_strlen(cachedName.get());
-  if (length < buflen) {
-    std::copy(cachedName.get(), cachedName.get() + length, buf);
-  } else {
-    length = 0;
-  }
-
-  buf[length] = '\0';
-  return true;
+  return result.append(cachedName.get(), js_strlen(cachedName.get()));
 }
 
-icu::TimeZone* js::DateTimeInfo::timeZone() {
-  if (!timeZone_) {
-    // The current default might be stale, because JS::ResetTimeZone()
-    // doesn't immediately update ICU's default time zone. So perform an
-    // update if needed.
-    js::ResyncICUDefaultTimeZone();
+static JS::UniqueChars DeflateString(mozilla::Span<const char16_t> chars) {
+  MOZ_ASSERT(mozilla::IsAscii(chars));
 
-    timeZone_.reset(icu::TimeZone::createDefault());
+  size_t length = chars.size();
+  JS::UniqueChars result(js_pod_malloc<char>(length + 1));
+  if (!result) {
+    return nullptr;
+  }
+
+  for (size_t i = 0; i < length; i++) {
+    result[i] = chars[i];
+  }
+  result[length] = '\0';
+
+  return result;
+}
+
+bool js::DateTimeInfo::internalTimeZoneId(TimeZoneIdentifierVector& result) {
+  if (!timeZoneId_) {
+    intl::FormatBuffer<char16_t,
+                       mozilla::intl::TimeZone::TimeZoneIdentifierLength,
+                       js::SystemAllocPolicy>
+        buffer;
+    if (timeZone()->GetId(buffer).isErr()) {
+      return false;
+    }
+
+    // ICU returns the time zone identifier as UTF-16, deflate to ASCII.
+    timeZoneId_ = DeflateString(buffer);
+    if (!timeZoneId_) {
+      return false;
+    }
+  }
+  return result.append(timeZoneId_.get(), js_strlen(timeZoneId_.get()));
+}
+
+mozilla::intl::TimeZone* js::DateTimeInfo::timeZone() {
+  if (!timeZone_) {
+    // For resist finger printing mode we always use the Atlantic/Reykjavik time
+    // zone as a "real world" UTC equivalent.
+    mozilla::Maybe<mozilla::Span<const char16_t>> timeZoneOverride;
+    if (forceUTC_) {
+      timeZoneOverride =
+          mozilla::Some(mozilla::MakeStringSpan(u"Atlantic/Reykjavik"));
+    }
+
+    auto timeZone = mozilla::intl::TimeZone::TryCreate(timeZoneOverride);
+
+    // Creating the default or UTC time zone should never fail. If it should
+    // fail nonetheless for some reason, just crash because we don't have a way
+    // to propagate any errors.
+    MOZ_RELEASE_ASSERT(timeZone.isOk());
+
+    timeZone_ = timeZone.unwrap();
     MOZ_ASSERT(timeZone_);
   }
 
   return timeZone_.get();
 }
-#endif /* ENABLE_INTL_API && !MOZ_SYSTEM_ICU */
+#endif /* JS_HAS_INTL_API */
 
 /* static */ js::ExclusiveData<js::DateTimeInfo>* js::DateTimeInfo::instance;
-
-/* static */ js::ExclusiveData<js::IcuTimeZoneStatus>* js::IcuTimeZoneState;
+/* static */ js::ExclusiveData<js::DateTimeInfo>* js::DateTimeInfo::instanceUTC;
 
 bool js::InitDateTimeState() {
-  MOZ_ASSERT(!DateTimeInfo::instance, "we should be initializing only once");
+  MOZ_ASSERT(!DateTimeInfo::instance && !DateTimeInfo::instanceUTC,
+             "we should be initializing only once");
 
   DateTimeInfo::instance =
-      js_new<ExclusiveData<DateTimeInfo>>(mutexid::DateTimeInfoMutex);
-  if (!DateTimeInfo::instance) {
-    return false;
-  }
-
-  MOZ_ASSERT(!IcuTimeZoneState, "we should be initializing only once");
-
-  // Set the ICU time zone status into the invalid state, so we compute the
-  // actual defaults on first access. We don't yet want to initialize ICU's
-  // time zone classes, because that may cause I/O operations slowing down
-  // the JS engine initialization, which we're currently in the middle of.
-  IcuTimeZoneState = js_new<ExclusiveData<IcuTimeZoneStatus>>(
-      mutexid::IcuTimeZoneStateMutex, IcuTimeZoneStatus::NeedsUpdate);
-  if (!IcuTimeZoneState) {
-    js_delete(DateTimeInfo::instance);
-    DateTimeInfo::instance = nullptr;
-    return false;
-  }
-
-  return true;
+      js_new<ExclusiveData<DateTimeInfo>>(mutexid::DateTimeInfoMutex, false);
+  DateTimeInfo::instanceUTC =
+      js_new<ExclusiveData<DateTimeInfo>>(mutexid::DateTimeInfoMutex, true);
+  return DateTimeInfo::instance && DateTimeInfo::instanceUTC;
 }
 
 /* static */
 void js::FinishDateTimeState() {
-  js_delete(IcuTimeZoneState);
-  IcuTimeZoneState = nullptr;
-
   js_delete(DateTimeInfo::instance);
   DateTimeInfo::instance = nullptr;
 }
 
 void js::ResetTimeZoneInternal(ResetTimeZoneMode mode) {
-  bool needsUpdate = js::DateTimeInfo::updateTimeZoneAdjustment(mode);
-
-#if ENABLE_INTL_API && defined(ICU_TZ_HAS_RECREATE_DEFAULT)
-  if (needsUpdate) {
-    auto guard = js::IcuTimeZoneState->lock();
-    guard.get() = js::IcuTimeZoneStatus::NeedsUpdate;
-  }
-#else
-  mozilla::Unused << needsUpdate;
-#endif
+  js::DateTimeInfo::resetTimeZone(mode);
 }
 
 JS_PUBLIC_API void JS::ResetTimeZone() {
-  js::ResetTimeZoneInternal(js::ResetTimeZoneMode::ResetEvenIfOffsetUnchaged);
+  js::ResetTimeZoneInternal(js::ResetTimeZoneMode::ResetEvenIfOffsetUnchanged);
 }
 
-#if defined(XP_WIN)
-static bool IsOlsonCompatibleWindowsTimeZoneId(const char* tz) {
+#if JS_HAS_INTL_API
+#  if defined(XP_WIN)
+static bool IsOlsonCompatibleWindowsTimeZoneId(std::string_view tz) {
   // ICU ignores the TZ environment variable on Windows and instead directly
   // invokes Win API functions to retrieve the current time zone. But since
   // we're still using the POSIX-derived localtime_s() function on Windows
@@ -589,23 +604,57 @@ static bool IsOlsonCompatibleWindowsTimeZoneId(const char* tz) {
       "GMT",
   };
   for (const auto& allowedId : allowedIds) {
-    if (std::strcmp(allowedId, tz) == 0) {
+    if (tz == allowedId) {
       return true;
     }
   }
   return false;
 }
-#elif ENABLE_INTL_API && defined(ICU_TZ_HAS_RECREATE_DEFAULT)
-static inline const char* TZContainsAbsolutePath(const char* tzVar) {
+#  else
+static std::string_view TZContainsAbsolutePath(std::string_view tzVar) {
   // A TZ environment variable may be an absolute path. The path
   // format of TZ may begin with a colon. (ICU handles relative paths.)
-  if (tzVar[0] == ':' && tzVar[1] == '/') {
-    return tzVar + 1;
+  if (tzVar.length() > 1 && tzVar[0] == ':' && tzVar[1] == '/') {
+    return tzVar.substr(1);
   }
-  if (tzVar[0] == '/') {
+  if (tzVar.length() > 0 && tzVar[0] == '/') {
     return tzVar;
   }
-  return nullptr;
+  return {};
+}
+
+/**
+ * Reject the input if it doesn't match the time zone id pattern or legacy time
+ * zone names.
+ *
+ * See <https://github.com/eggert/tz/blob/master/theory.html>.
+ */
+static bool IsTimeZoneId(std::string_view timeZone) {
+  size_t timeZoneLen = timeZone.length();
+
+  if (timeZoneLen == 0) {
+    return false;
+  }
+
+  for (size_t i = 0; i < timeZoneLen; i++) {
+    char c = timeZone[i];
+
+    // According to theory.html, '.' is allowed in time zone ids, but the
+    // accompanying zic.c file doesn't allow it. Assume the source file is
+    // correct and disallow '.' here, too.
+    if (mozilla::IsAsciiAlphanumeric(c) || c == '_' || c == '-' || c == '+') {
+      continue;
+    }
+
+    // Reject leading, trailing, or consecutive '/' characters.
+    if (c == '/' && i > 0 && i + 1 < timeZoneLen && timeZone[i + 1] != '/') {
+      continue;
+    }
+
+    return false;
+  }
+
+  return true;
 }
 
 /**
@@ -614,17 +663,23 @@ static inline const char* TZContainsAbsolutePath(const char* tzVar) {
  * path by repeatedly resolving symlinks until a path containing "/zoneinfo/"
  * followed by time zone looking components is found. If a symlink is broken,
  * symlink-following recurs too deeply, non time zone looking components are
- * encountered, or some other error is encountered, return the empty string.
+ * encountered, or some other error is encountered, then the |result| buffer is
+ * left empty.
  *
- * If a non-empty string is returned, it's only guaranteed to have certain
- * syntactic validity. It might not actually *be* a time zone name.
+ * If |result| is set to a non-empty string, it's only guaranteed to have
+ * certain syntactic validity. It might not actually *be* a time zone name.
+ *
+ * If there's an (OOM) error, |false| is returned.
  */
-static icu::UnicodeString ReadTimeZoneLink(const char* tz) {
+static bool ReadTimeZoneLink(std::string_view tz,
+                             js::TimeZoneIdentifierVector& result) {
+  MOZ_ASSERT(!tz.empty());
+  MOZ_ASSERT(result.empty());
+
   // The resolved link name can have different paths depending on the OS.
   // Follow ICU and only search for "/zoneinfo/"; see $ICU/common/putil.cpp.
   static constexpr char ZoneInfoPath[] = "/zoneinfo/";
-  constexpr size_t ZoneInfoPathLength =
-      mozilla::ArrayLength(ZoneInfoPath) - 1;  // exclude NUL
+  constexpr size_t ZoneInfoPathLength = js_strlen(ZoneInfoPath);
 
   // Stop following symlinks after a fixed depth, because some common time
   // zones are stored in files whose name doesn't match an Olson time zone
@@ -634,27 +689,28 @@ static icu::UnicodeString ReadTimeZoneLink(const char* tz) {
   // Four hops should be a reasonable limit for most use cases.
   constexpr uint32_t FollowDepthLimit = 4;
 
-#  ifdef PATH_MAX
+#    ifdef PATH_MAX
   constexpr size_t PathMax = PATH_MAX;
-#  else
+#    else
   constexpr size_t PathMax = 4096;
-#  endif
+#    endif
   static_assert(PathMax > 0, "PathMax should be larger than zero");
 
   char linkName[PathMax];
   constexpr size_t linkNameLen =
-      mozilla::ArrayLength(linkName) - 1;  // -1 to null-terminate.
+      std::size(linkName) - 1;  // -1 to null-terminate.
 
   // Return if the TZ value is too large.
-  if (std::strlen(tz) > linkNameLen) {
-    return icu::UnicodeString();
+  if (tz.length() > linkNameLen) {
+    return true;
   }
 
-  std::strcpy(linkName, tz);
+  tz.copy(linkName, tz.length());
+  linkName[tz.length()] = '\0';
 
   char linkTarget[PathMax];
   constexpr size_t linkTargetLen =
-      mozilla::ArrayLength(linkTarget) - 1;  // -1 to null-terminate.
+      std::size(linkTarget) - 1;  // -1 to null-terminate.
 
   uint32_t depth = 0;
 
@@ -663,13 +719,13 @@ static icu::UnicodeString ReadTimeZoneLink(const char* tz) {
   while (!(timeZoneWithZoneInfo = std::strstr(linkName, ZoneInfoPath))) {
     // Return if the symlink nesting is too deep.
     if (++depth > FollowDepthLimit) {
-      return icu::UnicodeString();
+      return true;
     }
 
     // Return on error or if the result was truncated.
     ssize_t slen = readlink(linkName, linkTarget, linkTargetLen);
     if (slen < 0 || size_t(slen) >= linkTargetLen) {
-      return icu::UnicodeString();
+      return true;
     }
 
     // Ensure linkTarget is null-terminated. (readlink may not necessarily
@@ -699,89 +755,93 @@ static icu::UnicodeString ReadTimeZoneLink(const char* tz) {
 
     // Return if the concatenated path name is too large.
     if (std::strlen(linkName) + len > linkNameLen) {
-      return icu::UnicodeString();
+      return true;
     }
 
     // Keep it simple and just concatenate the path names.
     std::strcat(linkName, linkTarget);
   }
 
-  const char* timeZone = timeZoneWithZoneInfo + ZoneInfoPathLength;
-  size_t timeZoneLen = std::strlen(timeZone);
+  std::string_view timeZone(timeZoneWithZoneInfo + ZoneInfoPathLength);
+  if (!IsTimeZoneId(timeZone)) {
+    return true;
+  }
+  return result.append(timeZone.data(), timeZone.length());
+}
+#  endif /* defined(XP_WIN) */
+#endif   /* JS_HAS_INTL_API */
 
-  // Reject the result if it doesn't match the time zone id pattern or
-  // legacy time zone names.
-  // See <https://github.com/eggert/tz/blob/master/theory.html>.
-  for (size_t i = 0; i < timeZoneLen; i++) {
-    char c = timeZone[i];
-
-    // According to theory.html, '.' is allowed in time zone ids, but the
-    // accompanying zic.c file doesn't allow it. Assume the source file is
-    // correct and disallow '.' here, too.
-    if (mozilla::IsAsciiAlphanumeric(c) || c == '_' || c == '-' || c == '+') {
-      continue;
-    }
-
-    // Reject leading, trailing, or consecutive '/' characters.
-    if (c == '/' && i > 0 && i + 1 < timeZoneLen && timeZone[i + 1] != '/') {
-      continue;
-    }
-
-    return icu::UnicodeString();
+void js::DateTimeInfo::internalResyncICUDefaultTimeZone() {
+#if JS_HAS_INTL_API
+  // In the future we should not be setting a default ICU time zone at all,
+  // instead all accesses should go through the appropriate DateTimeInfo
+  // instance depending on the resist fingerprinting status. For now we return
+  // early to prevent overwriting the default time zone with the UTC time zone
+  // used by RFP.
+  if (forceUTC_) {
+    return;
   }
 
-  return icu::UnicodeString(timeZone, timeZoneLen, US_INV);
-}
-#endif /* ENABLE_INTL_API && defined(ICU_TZ_HAS_RECREATE_DEFAULT) */
+  if (const char* tzenv = std::getenv("TZ")) {
+    std::string_view tz(tzenv);
 
-void js::ResyncICUDefaultTimeZone() {
-#if ENABLE_INTL_API && defined(ICU_TZ_HAS_RECREATE_DEFAULT)
-  auto guard = IcuTimeZoneState->lock();
-  if (guard.get() == IcuTimeZoneStatus::NeedsUpdate) {
-    bool recreate = true;
-
-    if (const char* tz = std::getenv("TZ")) {
-      icu::UnicodeString tzid;
+    mozilla::Span<const char> tzid;
 
 #  if defined(XP_WIN)
-      // If TZ is set and its value is valid under Windows' and IANA's
-      // time zone identifier rules, update the ICU default time zone to
-      // use this value.
-      if (IsOlsonCompatibleWindowsTimeZoneId(tz)) {
-        tzid.setTo(icu::UnicodeString(tz, -1, US_INV));
-      } else {
-        // If |tz| isn't a supported time zone identifier, use the
-        // default Windows time zone for ICU.
-        // TODO: Handle invalid time zone identifiers (bug 342068).
-      }
+    // If TZ is set and its value is valid under Windows' and IANA's time zone
+    // identifier rules, update the ICU default time zone to use this value.
+    if (IsOlsonCompatibleWindowsTimeZoneId(tz)) {
+      tzid = mozilla::Span(tz.data(), tz.length());
+    } else {
+      // If |tz| isn't a supported time zone identifier, use the default Windows
+      // time zone for ICU.
+      // TODO: Handle invalid time zone identifiers (bug 342068).
+    }
 #  else
-      // The TZ environment variable allows both absolute and
-      // relative paths, optionally beginning with a colon (':').
-      // (Relative paths, without the colon, are just Olson time
-      // zone names.)  We need to handle absolute paths ourselves,
-      // including handling that they might be symlinks.
-      // <https://unicode-org.atlassian.net/browse/ICU-13694>
-      if (const char* tzlink = TZContainsAbsolutePath(tz)) {
-        tzid.setTo(ReadTimeZoneLink(tzlink));
+    // The TZ environment variable allows both absolute and relative paths,
+    // optionally beginning with a colon (':'). (Relative paths, without the
+    // colon, are just Olson time zone names.)  We need to handle absolute paths
+    // ourselves, including handling that they might be symlinks.
+    // <https://unicode-org.atlassian.net/browse/ICU-13694>
+    TimeZoneIdentifierVector tzidVector;
+    std::string_view tzlink = TZContainsAbsolutePath(tz);
+    if (!tzlink.empty()) {
+      if (!ReadTimeZoneLink(tzlink, tzidVector)) {
+        // Ignore OOM.
+        return;
       }
+      tzid = tzidVector;
+    }
+
+#    ifdef ANDROID
+    // ICU ignores the TZ environment variable on Android. If it doesn't contain
+    // an absolute path, try to parse it as a time zone name.
+    else if (IsTimeZoneId(tz)) {
+      tzid = mozilla::Span(tz.data(), tz.length());
+    }
+#    endif
 #  endif /* defined(XP_WIN) */
 
-      if (!tzid.isEmpty()) {
-        mozilla::UniquePtr<icu::TimeZone> newTimeZone(
-            icu::TimeZone::createTimeZone(tzid));
-        MOZ_ASSERT(newTimeZone);
-        if (*newTimeZone != icu::TimeZone::getUnknown()) {
-          // adoptDefault() takes ownership of the time zone.
-          icu::TimeZone::adoptDefault(newTimeZone.release());
-          recreate = false;
-        }
+    if (!tzid.empty()) {
+      auto result = mozilla::intl::TimeZone::SetDefaultTimeZone(tzid);
+      if (result.isErr()) {
+        // Intentionally ignore any errors, because we don't have a good way to
+        // report errors from this function.
+        return;
       }
-    }
 
-    if (recreate) {
-      icu::TimeZone::recreateDefault();
+      // Return if the default time zone was successfully updated.
+      if (result.unwrap()) {
+        return;
+      }
+
+      // If SetDefaultTimeZone() succeeded, but the default time zone wasn't
+      // changed, proceed to set the default time zone from the host time zone.
     }
-    guard.get() = IcuTimeZoneStatus::Valid;
   }
+
+  // Intentionally ignore any errors, because we don't have a good way to report
+  // errors from this function.
+  (void)mozilla::intl::TimeZone::SetDefaultTimeZoneFromHostTimeZone();
 #endif
 }
