@@ -27,6 +27,10 @@ let types = {
   zip: "application/zip",
   json: "application/json",
   tar: "application/x-tar",
+  mp2: "audio/mpeg",
+  wav: "audio/x-wav",
+  m4a: "audio/mp4",
+  mp4: "video/mp4",
 };
 
 const PNG_DATA = atob(
