@@ -31,8 +31,9 @@ add_task(async function validate_filename_method() {
   Assert.equal(checkFilename("\\path.png", 0), "_path.png");
   Assert.equal(
     checkFilename("\\path*and/$?~file.png", 0),
-    "_path and_$ ~file.png"
+    "_path_and_$_~file.png"
   );
+
   Assert.equal(
     checkFilename(" \u180e whit\u180ee.png \u180e", 0),
     "whit\u180ee.png"
@@ -41,19 +42,19 @@ add_task(async function validate_filename_method() {
   Assert.equal(checkFilename(" happy\u061c\u2069.png", 0), "happy__.png");
   Assert.equal(
     checkFilename("12345678".repeat(31) + "abcdefgh.png", 0),
-    "12345678".repeat(31) + "abc.png"
+    "12345678".repeat(31) + "ab.png"
   );
   Assert.equal(
     checkFilename("簡単".repeat(41) + ".png", 0),
     "簡単".repeat(41) + ".png"
   );
   Assert.equal(
-    checkFilename("簡単".repeat(42) + ".png", 0),
-    "簡単".repeat(41) + "簡.png"
+    checkFilename("a" + "簡単".repeat(42) + ".png", 0),
+    "a" + "簡単".repeat(40) + "簡.png"
   );
   Assert.equal(
-    checkFilename("簡単".repeat(56) + ".png", 0),
-    "簡単".repeat(40) + "簡.png"
+    checkFilename("a" + "簡単".repeat(56) + ".png", 0),
+    "a" + "簡単".repeat(40) + ".png"
   );
   Assert.equal(checkFilename("café.png", 0), "café.png");
   Assert.equal(
@@ -62,7 +63,7 @@ add_task(async function validate_filename_method() {
   );
   Assert.equal(
     checkFilename("café".repeat(51) + ".png", 0),
-    "café".repeat(50) + ".png"
+    "café".repeat(49) + "caf.png"
   );
 
   Assert.equal(
@@ -84,11 +85,11 @@ add_task(async function validate_filename_method() {
   );
   Assert.equal(
     checkFilename("noextensionfile".repeat(17), 0),
-    "noextensionfile".repeat(16) + "noextension.png"
+    "noextensionfile".repeat(16) + "noextensio.png"
   );
   Assert.equal(
     checkFilename("noextensionfile".repeat(16) + "noextensionfil.", 0),
-    "noextensionfile".repeat(16) + "noextension.png"
+    "noextensionfile".repeat(16) + "noextensio.png"
   );
 
   Assert.equal(checkFilename("  first  .png  ", 0), "first .png");
@@ -103,12 +104,12 @@ add_task(async function validate_filename_method() {
   // For whatever reason, the Android mime handler accepts the .jpeg
   // extension for image/png, so skip this test there.
   if (AppConstants.platform != "android") {
-    Assert.equal(checkFilename("thi/*rd.jpeg", 0), "thi_ rd.png");
+    Assert.equal(checkFilename("thi/*rd.jpeg", 0), "thi__rd.png");
   }
 
   Assert.equal(
     checkFilename("f*\\ourth  file.jpg", mimeService.VALIDATE_SANITIZE_ONLY),
-    "f _ourth file.jpg"
+    "f__ourth file.jpg"
   );
   Assert.equal(
     checkFilename(
@@ -116,7 +117,7 @@ add_task(async function validate_filename_method() {
       mimeService.VALIDATE_SANITIZE_ONLY |
         mimeService.VALIDATE_DONT_COLLAPSE_WHITESPACE
     ),
-    "f _ift  h.jpe _g"
+    "f__ift  h.jpe__g"
   );
   Assert.equal(checkFilename("sixth.j  pe/*g", 0), "sixth.png");
 
@@ -130,13 +131,13 @@ add_task(async function validate_filename_method() {
   );
   Assert.equal(
     checkFilename(repeatStr + "seventh.png", 0),
-    repeatStr + "sev.png"
+    repeatStr + "se.png"
   );
 
   // no filename, so index is used by default.
-  Assert.equal(checkFilename(".png", 0), "index.png");
+  Assert.equal(checkFilename(".png", 0), "png.png");
 
-  // sanitization only, so index is not added, but initial period is stripped.
+  // sanitization only, so Untitled is not added, but initial period is stripped.
   Assert.equal(
     checkFilename(".png", mimeService.VALIDATE_SANITIZE_ONLY),
     "png"
@@ -154,28 +155,28 @@ add_task(async function validate_filename_method() {
   let ext = ".fairlyLongExtension";
   Assert.equal(
     checkFilename(repeatStr + ext, mimeService.VALIDATE_SANITIZE_ONLY),
-    repeatStr.substring(0, 255 - ext.length) + ext
+    repeatStr.substring(0, 254 - ext.length) + ext
   );
 
-  ext = "lo%?ng/invalid? ch\\ars";
+  ext = "lo#?n/ginvalid? ch\\ars";
   Assert.equal(
     checkFilename(repeatStr + ext, mimeService.VALIDATE_SANITIZE_ONLY),
-    repeatStr + "lo% ng_"
+    repeatStr + "lo#_n_"
   );
 
-  ext = ".long/invalid%? ch\\ars";
+  ext = ".long/invalid#? ch\\ars";
   Assert.equal(
     checkFilename(repeatStr + ext, mimeService.VALIDATE_SANITIZE_ONLY),
-    repeatStr.substring(0, 234) + ".long_invalid% ch_ars"
+    repeatStr.substring(0, 232) + ".long_invalid#_ch_ars"
   );
 
   Assert.equal(
     checkFilename("test_ﾃｽﾄ_T\x83E\\S\x83T.png", 0),
-    "test_ﾃｽﾄ_T E_S T.png"
+    "test_ﾃｽﾄ_T_E_S_T.png"
   );
   Assert.equal(
     checkFilename("test_ﾃｽﾄ_T\x83E\\S\x83T.pﾃ\x83ng", 0),
-    "test_ﾃｽﾄ_T E_S T.png"
+    "test_ﾃｽﾄ_T_E_S_T.png"
   );
 
   // Check we don't invalidate surrogate pairs when trimming.
@@ -188,15 +189,24 @@ add_task(async function validate_filename_method() {
     "video.ogg",
     "video.ogg"
   );
-  Assert.equal(
-    mimeService.validateFileNameForSaving("video.ogv", "video/ogg", 0),
+  let checkedName = mimeService.validateFileNameForSaving(
     "video.ogv",
-    "video.ogv"
+    "video/ogg",
+    0
   );
-  Assert.equal(
-    mimeService.validateFileNameForSaving("video.ogt", "video/ogg", 0),
-    "video.ogv",
-    "video.ogt"
+  Assert.ok(
+    ["video.ogv", "video.ogm"].includes(checkedName),
+    `Should get video.ogv or video.ogm for video.ogv, got ${checkedName}`
+  );
+
+  checkedName = mimeService.validateFileNameForSaving(
+    "video.ogt",
+    "video/ogg",
+    0
+  );
+  Assert.ok(
+    ["video.ogv", "video.ogm"].includes(checkedName),
+    `Should get video.ogv or video.ogm for video.ogt, got ${checkedName}`
   );
 
   Assert.equal(
@@ -215,16 +225,9 @@ add_task(async function validate_filename_method() {
     "audio.mp2"
   );
 
-  let expected = "audio.mp3";
-  if (AppConstants.platform == "linux") {
-    expected = "audio.mpga";
-  } else if (AppConstants.platform == "android") {
-    expected = "audio.mp4";
-  }
-
   Assert.equal(
     mimeService.validateFileNameForSaving("audio.mp4", "audio/mpeg", 0),
-    expected,
+    AppConstants.platform == "android" ? "audio.mp4" : "audio.mp3",
     "audio.mp4"
   );
 
@@ -248,11 +251,11 @@ add_task(async function validate_filename_method() {
   // cropped to fit into 255 bytes.
   Assert.equal(
     mimeService.validateFileNameForSaving(
-      "라이브9.9만 시청컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장24%102 000원 브랜드데이 앵콜 🎁 1.등 유산균 컬처렐 특가!",
+      "라이브9.9만 시청컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장24%102 000원 브랜드데이 앵콜 🎁 1.등-유산균-컬처렐-특가!",
       "text/unknown",
       mimeService.VALIDATE_SANITIZE_ONLY
     ),
-    "라이브9.9만 시청컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 .등 유산균 컬처렐 특가!",
+    "라이브9.9만 시청컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 .등-유산균-컬처렐-특가!",
     "very long filename with extension"
   );
 
@@ -270,11 +273,145 @@ add_task(async function validate_filename_method() {
   // This filename is cropped at 254 bytes.
   Assert.equal(
     mimeService.validateFileNameForSaving(
-      ".라이브99만 시청컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장24%102 000원 브랜드데이 앵콜 🎁 1등 유산균 컬처렐 특가!",
+      ".라이브99만 시청컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장24_102 000원 브랜드데이 앵콜 🎁 1등 유산균 컬처렐 특가!",
       "text/unknown",
       mimeService.VALIDATE_SANITIZE_ONLY
     ),
-    "라이브99만 시청컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장24%102 000원 브랜드데",
+    "라이브99만 시청컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장컬처렐 다이제스티브 3박스 - 3박스 더 (뚱랑이 굿즈 증정) - 선물용 쇼핑백 2장24_102 000원 브랜드데",
     "very filename with extension only"
   );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving("1. First", "text/unknown", 0),
+    "1.First",
+    "1. First"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving(
+      "1. First",
+      "text/unknown",
+      mimeService.VALIDATE_ALLOW_DIRECTORY_NAMES
+    ),
+    "1. First",
+    "1. First with VALIDATE_ALLOW_DIRECTORY_NAMES"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving("filename.LNK", "text/unknown", 0),
+    "filename.LNK.download",
+    "filename.LNK"
+  );
+  Assert.equal(
+    mimeService.validateFileNameForSaving("filename.local", "text/unknown", 0),
+    "filename.local.download",
+    "filename.local"
+  );
+  Assert.equal(
+    mimeService.validateFileNameForSaving("filename.url", "text/unknown", 0),
+    "filename.url.download",
+    "filename.url"
+  );
+  Assert.equal(
+    mimeService.validateFileNameForSaving("filename.URl", "text/unknown", 0),
+    "filename.URl.download",
+    "filename.URl"
+  );
+  Assert.equal(
+    mimeService.validateFileNameForSaving("filename.scf", "text/unknown", 0),
+    "filename.scf.download",
+    "filename.scf"
+  );
+  Assert.equal(
+    mimeService.validateFileNameForSaving("filename.sCF", "text/unknown", 0),
+    "filename.sCF.download",
+    "filename.sCF"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving("filename.lnk\n", "text/unknown", 0),
+    "filename.lnk_",
+    "filename.lnk with newline"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving(
+      "filename.lnk\n  ",
+      "text/unknown",
+      0
+    ),
+    "filename.lnk_",
+    "filename.lnk with newline"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving(
+      "filename.\n\t  lnk",
+      "text/unknown",
+      0
+    ),
+    "filename.__lnk",
+    "filename.lnk with space and newline"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving(
+      "filename.local\u180e\u180e\u180e",
+      "text/unknown",
+      0
+    ),
+    "filename.local.download",
+    "filename.lnk with vowel separators"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving(
+      "filename.LNK",
+      "text/unknown",
+      mimeService.VALIDATE_SANITIZE_ONLY
+    ),
+    "filename.LNK.download",
+    "filename.LNK sanitize only"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving(
+      "filename.LNK\n",
+      "text/unknown",
+      mimeService.VALIDATE_ALLOW_INVALID_FILENAMES
+    ),
+    "filename.LNK_",
+    "filename.LNK allow invalid"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving(
+      "filename.URL\n",
+      "text/unknown",
+      mimeService.VALIDATE_SANITIZE_ONLY |
+        mimeService.VALIDATE_ALLOW_INVALID_FILENAMES
+    ),
+    "filename.URL_",
+    "filename.URL allow invalid, sanitize only"
+  );
+
+  Assert.equal(
+    mimeService.validateFileNameForSaving(
+      "filename.desktop",
+      "text/unknown",
+      mimeService.VALIDATE_SANITIZE_ONLY
+    ),
+    "filename.desktop.download",
+    "filename.desktop sanitize only"
+  );
+  Assert.equal(
+    mimeService.validateFileNameForSaving(
+      "filename.DESKTOP\n",
+      "text/unknown",
+      mimeService.VALIDATE_SANITIZE_ONLY |
+        mimeService.VALIDATE_ALLOW_INVALID_FILENAMES
+    ),
+    "filename.DESKTOP_",
+    "filename.DESKTOP allow invalid, sanitize only"
+  );
 });
