@@ -7,25 +7,24 @@
 
 #include "base/basictypes.h"
 
-#include "ClientLayerManager.h"
 #include "gfxPlatform.h"
 #include "nsRefreshDriver.h"
 #include "mozilla/dom/BrowserChild.h"
+#include "mozilla/EventForwards.h"
 #include "mozilla/gfx/gfxVars.h"
-#include "mozilla/Hal.h"
 #include "mozilla/IMEStateManager.h"
 #include "mozilla/layers/APZChild.h"
-#include "mozilla/layers/PLayerTransactionChild.h"
 #include "mozilla/layers/WebRenderLayerManager.h"
+#include "mozilla/NativeKeyBindingsType.h"
 #include "mozilla/Preferences.h"
 #include "mozilla/PresShell.h"
 #include "mozilla/SchedulerGroup.h"
 #include "mozilla/StaticPrefs_browser.h"
+#include "mozilla/StaticPrefs_gfx.h"
 #include "mozilla/TextComposition.h"
 #include "mozilla/TextEventDispatcher.h"
 #include "mozilla/TextEvents.h"
 #include "mozilla/Unused.h"
-#include "BasicLayers.h"
 #include "PuppetWidget.h"
 #include "nsContentUtils.h"
 #include "nsIWidgetListener.h"
@@ -36,7 +35,6 @@
 
 using namespace mozilla;
 using namespace mozilla::dom;
-using namespace mozilla::hal;
 using namespace mozilla::gfx;
 using namespace mozilla::layers;
 using namespace mozilla::widget;
@@ -61,13 +59,15 @@ already_AddRefed<nsIWidget> nsIWidget::CreatePuppetWidget(
 namespace mozilla {
 namespace widget {
 
-static bool IsPopup(const nsWidgetInitData* aInitData) {
-  return aInitData && aInitData->mWindowType == eWindowType_popup;
+static bool IsPopup(const widget::InitData* aInitData) {
+  return aInitData && aInitData->mWindowType == WindowType::Popup;
 }
 
-static bool MightNeedIMEFocus(const nsWidgetInitData* aInitData) {
+static bool MightNeedIMEFocus(const widget::InitData* aInitData) {
   // In the puppet-widget world, popup widgets are just dummies and
   // shouldn't try to mess with IME state.
+  //
+  // FIXME(emilio): Pretty sure we no longer have puppet popups.
 #ifdef MOZ_CROSS_PROCESS_IME
   return !IsPopup(aInitData);
 #else
@@ -75,22 +75,18 @@ static bool MightNeedIMEFocus(const nsWidgetInitData* aInitData) {
 #endif
 }
 
-// Arbitrary, fungible.
-const size_t PuppetWidget::kMaxDimension = 4000;
-
 NS_IMPL_ISUPPORTS_INHERITED(PuppetWidget, nsBaseWidget,
                             TextEventDispatcherListener)
 
 PuppetWidget::PuppetWidget(BrowserChild* aBrowserChild)
     : mBrowserChild(aBrowserChild),
       mMemoryPressureObserver(nullptr),
-      mDPI(-1),
-      mRounding(1),
-      mDefaultScale(-1),
       mEnabled(false),
       mVisible(false),
+      mSizeMode(nsSizeMode_Normal),
       mNeedIMEStateInit(false),
       mIgnoreCompositionEvents(false) {
+  mWidgetType = WidgetType::Puppet;
   // Setting 'Unknown' means "not yet cached".
   mInputContext.mIMEState.mEnabled = IMEEnabled::Unknown;
 }
@@ -98,36 +94,24 @@ PuppetWidget::PuppetWidget(BrowserChild* aBrowserChild)
 PuppetWidget::~PuppetWidget() { Destroy(); }
 
 void PuppetWidget::InfallibleCreate(nsIWidget* aParent,
-                                    nsNativeWidget aNativeParent,
                                     const LayoutDeviceIntRect& aRect,
-                                    nsWidgetInitData* aInitData) {
-  MOZ_ASSERT(!aNativeParent, "got a non-Puppet native parent");
-
-  BaseCreate(nullptr, aInitData);
+                                    widget::InitData* aInitData) {
+  BaseCreate(aParent, aInitData);
 
   mBounds = aRect;
   mEnabled = true;
   mVisible = true;
 
-  mDrawTarget = gfxPlatform::GetPlatform()->CreateOffscreenContentDrawTarget(
-      IntSize(1, 1), SurfaceFormat::B8G8R8A8);
-
   mNeedIMEStateInit = MightNeedIMEFocus(aInitData);
 
-  PuppetWidget* parent = static_cast<PuppetWidget*>(aParent);
-  if (parent) {
-    parent->SetChild(this);
-    mLayerManager = parent->GetLayerManager();
-  } else {
-    Resize(mBounds.X(), mBounds.Y(), mBounds.Width(), mBounds.Height(), false);
-  }
+  Resize(mBounds.X(), mBounds.Y(), mBounds.Width(), mBounds.Height(), false);
   mMemoryPressureObserver = MemoryPressureObserver::Create(this);
 }
 
-nsresult PuppetWidget::Create(nsIWidget* aParent, nsNativeWidget aNativeParent,
+nsresult PuppetWidget::Create(nsIWidget* aParent,
                               const LayoutDeviceIntRect& aRect,
-                              nsWidgetInitData* aInitData) {
-  InfallibleCreate(aParent, aNativeParent, aRect, aInitData);
+                              widget::InitData* aInitData) {
+  InfallibleCreate(aParent, aRect, aInitData);
   return NS_OK;
 }
 
@@ -141,17 +125,6 @@ void PuppetWidget::InitIMEState() {
   }
 }
 
-already_AddRefed<nsIWidget> PuppetWidget::CreateChild(
-    const LayoutDeviceIntRect& aRect, nsWidgetInitData* aInitData,
-    bool aForceUseIWidgetParent) {
-  bool isPopup = IsPopup(aInitData);
-  nsCOMPtr<nsIWidget> widget = nsIWidget::CreatePuppetWidget(mBrowserChild);
-  return ((widget && NS_SUCCEEDED(widget->Create(isPopup ? nullptr : this,
-                                                 nullptr, aRect, aInitData)))
-              ? widget.forget()
-              : nullptr);
-}
-
 void PuppetWidget::Destroy() {
   if (mOnDestroyCalled) {
     return;
@@ -164,11 +137,10 @@ void PuppetWidget::Destroy() {
     mMemoryPressureObserver->Unregister();
     mMemoryPressureObserver = nullptr;
   }
-  mChild = nullptr;
-  if (mLayerManager) {
-    mLayerManager->Destroy();
+  if (mWindowRenderer) {
+    mWindowRenderer->Destroy();
   }
-  mLayerManager = nullptr;
+  mWindowRenderer = nullptr;
   mBrowserChild = nullptr;
 }
 
@@ -179,10 +151,6 @@ void PuppetWidget::Show(bool aState) {
   bool wasVisible = mVisible;
   mVisible = aState;
 
-  if (mChild) {
-    mChild->mVisible = aState;
-  }
-
   if (!wasVisible && mVisible) {
     // The previously attached widget listener is handy if
     // we're transitioning from page to page without dropping
@@ -203,11 +171,6 @@ void PuppetWidget::Resize(double aWidth, double aHeight, bool aRepaint) {
   mBounds.SizeTo(
       LayoutDeviceIntSize(NSToIntRound(aWidth), NSToIntRound(aHeight)));
 
-  if (mChild) {
-    mChild->Resize(aWidth, aHeight, aRepaint);
-    return;
-  }
-
   // XXX: roc says that |aRepaint| dictates whether or not to
   // invalidate the expanded area
   if (oldBounds.Size() < mBounds.Size() && aRepaint) {
@@ -230,26 +193,6 @@ void PuppetWidget::Resize(double aWidth, double aHeight, bool aRepaint) {
   }
 }
 
-nsresult PuppetWidget::ConfigureChildren(
-    const nsTArray<Configuration>& aConfigurations) {
-  for (uint32_t i = 0; i < aConfigurations.Length(); ++i) {
-    const Configuration& configuration = aConfigurations[i];
-    PuppetWidget* w = static_cast<PuppetWidget*>(configuration.mChild.get());
-    NS_ASSERTION(w->GetParent() == this, "Configured widget is not a child");
-    w->SetWindowClipRegion(configuration.mClipRegion, true);
-    LayoutDeviceIntRect bounds = w->GetBounds();
-    if (bounds.Size() != configuration.mBounds.Size()) {
-      w->Resize(configuration.mBounds.X(), configuration.mBounds.Y(),
-                configuration.mBounds.Width(), configuration.mBounds.Height(),
-                true);
-    } else if (bounds.TopLeft() != configuration.mBounds.TopLeft()) {
-      w->Move(configuration.mBounds.X(), configuration.mBounds.Y());
-    }
-    w->SetWindowClipRegion(configuration.mClipRegion, false);
-  }
-  return NS_OK;
-}
-
 void PuppetWidget::SetFocus(Raise aRaise, CallerType aCallerType) {
   if (aRaise == Raise::Yes && mBrowserChild) {
     mBrowserChild->SendRequestFocus(true, aCallerType);
@@ -261,25 +204,19 @@ void PuppetWidget::Invalidate(const LayoutDeviceIntRect& aRect) {
   debug_DumpInvalidate(stderr, this, &aRect, "PuppetWidget", 0);
 #endif
 
-  if (mChild) {
-    mChild->Invalidate(aRect);
-    return;
-  }
-
   if (mBrowserChild && !aRect.IsEmpty() && !mWidgetPaintTask.IsPending()) {
     mWidgetPaintTask = new WidgetPaintTask(this);
     nsCOMPtr<nsIRunnable> event(mWidgetPaintTask.get());
-    SchedulerGroup::Dispatch(TaskCategory::Other, event.forget());
+    SchedulerGroup::Dispatch(event.forget());
   }
 }
 
 mozilla::LayoutDeviceToLayoutDeviceMatrix4x4
 PuppetWidget::WidgetToTopLevelWidgetTransform() {
-  if (!GetOwningBrowserChild()) {
-    NS_WARNING("PuppetWidget without Tab does not have transform information.");
-    return mozilla::LayoutDeviceToLayoutDeviceMatrix4x4();
+  if (auto* bc = GetOwningBrowserChild()) {
+    return bc->GetChildToParentConversionMatrix();
   }
-  return GetOwningBrowserChild()->GetChildToParentConversionMatrix();
+  return mozilla::LayoutDeviceToLayoutDeviceMatrix4x4();
 }
 
 void PuppetWidget::InitEvent(WidgetGUIEvent& aEvent,
@@ -290,7 +227,6 @@ void PuppetWidget::InitEvent(WidgetGUIEvent& aEvent,
     // use the point override if provided
     aEvent.mRefPoint = *aPoint;
   }
-  aEvent.mTime = PR_Now() / 1000;
 }
 
 nsresult PuppetWidget::DispatchEvent(WidgetGUIEvent* aEvent,
@@ -299,9 +235,6 @@ nsresult PuppetWidget::DispatchEvent(WidgetGUIEvent* aEvent,
   debug_DumpEvent(stdout, aEvent->mWidget, aEvent, "PuppetWidget", 0);
 #endif
 
-  MOZ_ASSERT(!mChild || mChild->mWindowType == eWindowType_popup,
-             "Unexpected event dispatch!");
-
   MOZ_ASSERT(!aEvent->AsKeyboardEvent() ||
                  aEvent->mFlags.mIsSynthesizedForTests ||
                  aEvent->AsKeyboardEvent()->AreAllEditCommandsInitialized(),
@@ -392,6 +325,9 @@ nsIWidget::ContentAndAPZEventStatus PuppetWidget::DispatchInputEvent(
     return status;
   }
 
+  MOZ_ASSERT(aEvent->mMessage != ePointerClick);
+  MOZ_ASSERT(aEvent->mMessage != ePointerAuxClick);
+
   switch (aEvent->mClass) {
     case eWheelEventClass:
       Unused << mBrowserChild->SendDispatchWheelEvent(*aEvent->AsWheelEvent());
@@ -416,73 +352,74 @@ nsIWidget::ContentAndAPZEventStatus PuppetWidget::DispatchInputEvent(
 nsresult PuppetWidget::SynthesizeNativeKeyEvent(
     int32_t aNativeKeyboardLayout, int32_t aNativeKeyCode,
     uint32_t aModifierFlags, const nsAString& aCharacters,
-    const nsAString& aUnmodifiedCharacters, nsIObserver* aObserver) {
-  AutoObserverNotifier notifier(aObserver, "keyevent");
+    const nsAString& aUnmodifiedCharacters,
+    nsISynthesizedEventCallback* aCallback) {
+  AutoSynthesizedEventCallbackNotifier notifier(aCallback);
   if (!mBrowserChild) {
     return NS_ERROR_FAILURE;
   }
   mBrowserChild->SendSynthesizeNativeKeyEvent(
-      aNativeKeyboardLayout, aNativeKeyCode, aModifierFlags,
-      nsString(aCharacters), nsString(aUnmodifiedCharacters),
-      notifier.SaveObserver());
+      aNativeKeyboardLayout, aNativeKeyCode, aModifierFlags, aCharacters,
+      aUnmodifiedCharacters, notifier.SaveCallback());
   return NS_OK;
 }
 
 nsresult PuppetWidget::SynthesizeNativeMouseEvent(
     mozilla::LayoutDeviceIntPoint aPoint, NativeMouseMessage aNativeMessage,
     MouseButton aButton, nsIWidget::Modifiers aModifierFlags,
-    nsIObserver* aObserver) {
-  AutoObserverNotifier notifier(aObserver, "mouseevent");
+    nsISynthesizedEventCallback* aCallback) {
+  AutoSynthesizedEventCallbackNotifier notifier(aCallback);
   if (!mBrowserChild) {
     return NS_ERROR_FAILURE;
   }
   mBrowserChild->SendSynthesizeNativeMouseEvent(
       aPoint, static_cast<uint32_t>(aNativeMessage),
       static_cast<int16_t>(aButton), static_cast<uint32_t>(aModifierFlags),
-      notifier.SaveObserver());
+      notifier.SaveCallback());
   return NS_OK;
 }
 
 nsresult PuppetWidget::SynthesizeNativeMouseMove(
-    mozilla::LayoutDeviceIntPoint aPoint, nsIObserver* aObserver) {
-  AutoObserverNotifier notifier(aObserver, "mousemove");
+    mozilla::LayoutDeviceIntPoint aPoint,
+    nsISynthesizedEventCallback* aCallback) {
+  AutoSynthesizedEventCallbackNotifier notifier(aCallback);
   if (!mBrowserChild) {
     return NS_ERROR_FAILURE;
   }
-  mBrowserChild->SendSynthesizeNativeMouseMove(aPoint, notifier.SaveObserver());
+  mBrowserChild->SendSynthesizeNativeMouseMove(aPoint, notifier.SaveCallback());
   return NS_OK;
 }
 
 nsresult PuppetWidget::SynthesizeNativeMouseScrollEvent(
     mozilla::LayoutDeviceIntPoint aPoint, uint32_t aNativeMessage,
     double aDeltaX, double aDeltaY, double aDeltaZ, uint32_t aModifierFlags,
-    uint32_t aAdditionalFlags, nsIObserver* aObserver) {
-  AutoObserverNotifier notifier(aObserver, "mousescrollevent");
+    uint32_t aAdditionalFlags, nsISynthesizedEventCallback* aCallback) {
+  AutoSynthesizedEventCallbackNotifier notifier(aCallback);
   if (!mBrowserChild) {
     return NS_ERROR_FAILURE;
   }
   mBrowserChild->SendSynthesizeNativeMouseScrollEvent(
       aPoint, aNativeMessage, aDeltaX, aDeltaY, aDeltaZ, aModifierFlags,
-      aAdditionalFlags, notifier.SaveObserver());
+      aAdditionalFlags, notifier.SaveCallback());
   return NS_OK;
 }
 
 nsresult PuppetWidget::SynthesizeNativeTouchPoint(
     uint32_t aPointerId, TouchPointerState aPointerState,
     LayoutDeviceIntPoint aPoint, double aPointerPressure,
-    uint32_t aPointerOrientation, nsIObserver* aObserver) {
-  AutoObserverNotifier notifier(aObserver, "touchpoint");
+    uint32_t aPointerOrientation, nsISynthesizedEventCallback* aCallback) {
+  AutoSynthesizedEventCallbackNotifier notifier(aCallback);
   if (!mBrowserChild) {
     return NS_ERROR_FAILURE;
   }
   mBrowserChild->SendSynthesizeNativeTouchPoint(
       aPointerId, aPointerState, aPoint, aPointerPressure, aPointerOrientation,
-      notifier.SaveObserver());
+      notifier.SaveCallback());
   return NS_OK;
 }
 
 nsresult PuppetWidget::SynthesizeNativeTouchPadPinch(
-    TouchpadPinchPhase aEventPhase, float aScale, LayoutDeviceIntPoint aPoint,
+    TouchpadGesturePhase aEventPhase, float aScale, LayoutDeviceIntPoint aPoint,
     int32_t aModifierFlags) {
   if (!mBrowserChild) {
     return NS_ERROR_FAILURE;
@@ -492,38 +429,30 @@ nsresult PuppetWidget::SynthesizeNativeTouchPadPinch(
   return NS_OK;
 }
 
-nsresult PuppetWidget::SynthesizeNativeTouchTap(LayoutDeviceIntPoint aPoint,
-                                                bool aLongTap,
-                                                nsIObserver* aObserver) {
-  AutoObserverNotifier notifier(aObserver, "touchtap");
+nsresult PuppetWidget::SynthesizeNativeTouchTap(
+    LayoutDeviceIntPoint aPoint, bool aLongTap,
+    nsISynthesizedEventCallback* aCallback) {
+  AutoSynthesizedEventCallbackNotifier notifier(aCallback);
   if (!mBrowserChild) {
     return NS_ERROR_FAILURE;
   }
   mBrowserChild->SendSynthesizeNativeTouchTap(aPoint, aLongTap,
-                                              notifier.SaveObserver());
-  return NS_OK;
-}
-
-nsresult PuppetWidget::ClearNativeTouchSequence(nsIObserver* aObserver) {
-  AutoObserverNotifier notifier(aObserver, "cleartouch");
-  if (!mBrowserChild) {
-    return NS_ERROR_FAILURE;
-  }
-  mBrowserChild->SendClearNativeTouchSequence(notifier.SaveObserver());
+                                              notifier.SaveCallback());
   return NS_OK;
 }
 
 nsresult PuppetWidget::SynthesizeNativePenInput(
     uint32_t aPointerId, TouchPointerState aPointerState,
     LayoutDeviceIntPoint aPoint, double aPressure, uint32_t aRotation,
-    int32_t aTiltX, int32_t aTiltY, nsIObserver* aObserver) {
-  AutoObserverNotifier notifier(aObserver, "peninput");
+    int32_t aTiltX, int32_t aTiltY, int32_t aButton,
+    nsISynthesizedEventCallback* aCallback) {
+  AutoSynthesizedEventCallbackNotifier notifier(aCallback);
   if (!mBrowserChild) {
     return NS_ERROR_FAILURE;
   }
-  mBrowserChild->SendSynthesizeNativePenInput(aPointerId, aPointerState, aPoint,
-                                              aPressure, aRotation, aTiltX,
-                                              aTiltY, notifier.SaveObserver());
+  mBrowserChild->SendSynthesizeNativePenInput(
+      aPointerId, aPointerState, aPoint, aPressure, aRotation, aTiltX, aTiltY,
+      aButton, notifier.SaveCallback());
   return NS_OK;
 }
 
@@ -536,6 +465,20 @@ nsresult PuppetWidget::SynthesizeNativeTouchpadDoubleTap(
   return NS_OK;
 }
 
+nsresult PuppetWidget::SynthesizeNativeTouchpadPan(
+    TouchpadGesturePhase aEventPhase, LayoutDeviceIntPoint aPoint,
+    double aDeltaX, double aDeltaY, int32_t aModifierFlags,
+    nsISynthesizedEventCallback* aCallback) {
+  AutoSynthesizedEventCallbackNotifier notifier(aCallback);
+  if (!mBrowserChild) {
+    return NS_ERROR_FAILURE;
+  }
+  mBrowserChild->SendSynthesizeNativeTouchpadPan(aEventPhase, aPoint, aDeltaX,
+                                                 aDeltaY, aModifierFlags,
+                                                 notifier.SaveCallback());
+  return NS_OK;
+}
+
 void PuppetWidget::LockNativePointer() {
   if (!mBrowserChild) {
     return;
@@ -585,16 +528,13 @@ bool PuppetWidget::GetEditCommands(NativeKeyBindingsType aType,
   return true;
 }
 
-LayerManager* PuppetWidget::GetLayerManager(
-    PLayerTransactionChild* aShadowManager, LayersBackend aBackendHint,
-    LayerManagerPersistence aPersistence) {
-  if (!mLayerManager) {
+WindowRenderer* PuppetWidget::GetWindowRenderer() {
+  if (!mWindowRenderer) {
     if (XRE_IsParentProcess()) {
       // On the parent process there is no CompositorBridgeChild which confuses
       // some layers code, so we use basic layers instead. Note that we create
-      // a non-retaining layer manager since we don't care about performance.
-      mLayerManager = new BasicLayerManager(BasicLayerManager::BLM_OFFSCREEN);
-      return mLayerManager;
+      mWindowRenderer = new FallbackRenderer;
+      return mWindowRenderer;
     }
 
     // If we know for sure that the parent side of this BrowserChild is not
@@ -603,21 +543,16 @@ LayerManager* PuppetWidget::GetLayerManager(
     // can do drawing in this process.
     MOZ_ASSERT(!mBrowserChild ||
                mBrowserChild->IsLayersConnected() != Some(true));
-    mLayerManager = new BasicLayerManager(this);
+    mWindowRenderer = CreateFallbackRenderer();
   }
 
-  return mLayerManager;
+  return mWindowRenderer;
 }
 
 bool PuppetWidget::CreateRemoteLayerManager(
-    const std::function<bool(LayerManager*)>& aInitializeFunc) {
-  RefPtr<LayerManager> lm;
+    const std::function<bool(WebRenderLayerManager*)>& aInitializeFunc) {
+  RefPtr<WebRenderLayerManager> lm = new WebRenderLayerManager(this);
   MOZ_ASSERT(mBrowserChild);
-  if (mBrowserChild->GetCompositorOptions().UseWebRender()) {
-    lm = new WebRenderLayerManager(this);
-  } else {
-    lm = new ClientLayerManager(this);
-  }
 
   if (!aInitializeFunc(lm)) {
     return false;
@@ -628,7 +563,7 @@ bool PuppetWidget::CreateRemoteLayerManager(
   // it if we successfully create its successor because a partially initialized
   // layer manager is worse than a fully initialized but shutdown layer manager.
   DestroyLayerManager();
-  mLayerManager = std::move(lm);
+  mWindowRenderer = std::move(lm);
   return true;
 }
 
@@ -669,7 +604,7 @@ nsresult PuppetWidget::RequestIMEToCommitComposition(bool aCancel) {
   bool isCommitted = false;
   nsAutoString committedString;
   if (NS_WARN_IF(!mBrowserChild->SendRequestIMEToCommitComposition(
-          aCancel, &isCommitted, &committedString))) {
+          aCancel, composition->Id(), &isCommitted, &committedString))) {
     return NS_ERROR_FAILURE;
   }
 
@@ -697,7 +632,7 @@ nsresult PuppetWidget::RequestIMEToCommitComposition(bool aCancel) {
   mIgnoreCompositionEvents = true;
 
   Unused << mBrowserChild->SendOnEventNeedingAckHandled(
-      eCompositionCommitRequestHandled);
+      eCompositionCommitRequestHandled, composition->Id());
 
   // NOTE: PuppetWidget might be destroyed already.
   return NS_OK;
@@ -772,7 +707,8 @@ nsresult PuppetWidget::NotifyIMEOfFocusChange(
   bool gotFocus = aIMENotification.mMessage == NOTIFY_IME_OF_FOCUS;
   if (gotFocus) {
     // When IME gets focus, we should initialize all information of the
-    // content.
+    // content, however, it may fail to get it because the editor may have
+    // already been blurred.
     if (NS_WARN_IF(!mContentCache.CacheAll(this, &aIMENotification))) {
       return NS_ERROR_FAILURE;
     }
@@ -809,7 +745,8 @@ nsresult PuppetWidget::NotifyIMEOfCompositionUpdate(
     return NS_ERROR_FAILURE;
   }
 
-  if (NS_WARN_IF(!mContentCache.CacheSelection(this, &aIMENotification))) {
+  if (NS_WARN_IF(
+          !mContentCache.CacheCaretAndTextRects(this, &aIMENotification))) {
     return NS_ERROR_FAILURE;
   }
   mBrowserChild->SendNotifyIMECompositionUpdate(mContentCache,
@@ -856,11 +793,13 @@ nsresult PuppetWidget::NotifyIMEOfSelectionChange(
 
   // Note that selection change must be notified after text change if it occurs.
   // Therefore, we don't need to query text content again here.
-  mContentCache.SetSelection(
-      this, aIMENotification.mSelectionChangeData.mOffset,
-      aIMENotification.mSelectionChangeData.Length(),
-      aIMENotification.mSelectionChangeData.mReversed,
-      aIMENotification.mSelectionChangeData.GetWritingMode());
+  if (MOZ_UNLIKELY(!mContentCache.SetSelection(
+          this, aIMENotification.mSelectionChangeData))) {
+    // If there is no text cache yet, caching text will cache selection too.
+    // Therefore, in the case, we don't need to notify IME of selection change
+    // right now.
+    return NS_OK;
+  }
 
   mBrowserChild->SendNotifyIMESelection(mContentCache, aIMENotification);
 
@@ -893,7 +832,8 @@ nsresult PuppetWidget::NotifyIMEOfPositionChange(
   if (NS_WARN_IF(!mContentCache.CacheEditorRect(this, &aIMENotification))) {
     return NS_ERROR_FAILURE;
   }
-  if (NS_WARN_IF(!mContentCache.CacheSelection(this, &aIMENotification))) {
+  if (NS_WARN_IF(
+          !mContentCache.CacheCaretAndTextRects(this, &aIMENotification))) {
     return NS_ERROR_FAILURE;
   }
   if (mIMENotificationRequestsOfParent.WantPositionChanged()) {
@@ -919,17 +859,14 @@ void PuppetWidget::SetCursor(const Cursor& aCursor) {
     return;
   }
 
-  bool hasCustomCursor = false;
-  UniquePtr<char[]> customCursorData;
-  size_t length = 0;
-  IntSize customCursorSize;
-  int32_t stride = 0;
-  auto format = SurfaceFormat::B8G8R8A8;
   ImageResolution resolution = aCursor.mResolution;
+  Maybe<IPCImage> customCursor;
   if (aCursor.IsCustom()) {
-    int32_t width = 0, height = 0;
+    int32_t width = 0;
+    int32_t height = 0;
     aCursor.mContainer->GetWidth(&width);
     aCursor.mContainer->GetHeight(&height);
+
     const int32_t flags =
         imgIContainer::FLAG_SYNC_DECODE | imgIContainer::FLAG_ASYNC_NOTIFY;
     RefPtr<SourceSurface> surface;
@@ -947,35 +884,21 @@ void PuppetWidget::SetCursor(const Cursor& aCursor) {
       surface =
           aCursor.mContainer->GetFrame(imgIContainer::FRAME_CURRENT, flags);
     }
+
     if (surface) {
       if (RefPtr<DataSourceSurface> dataSurface = surface->GetDataSurface()) {
-        hasCustomCursor = true;
-        customCursorData = nsContentUtils::GetSurfaceData(
-            WrapNotNull(dataSurface), &length, &stride);
-        customCursorSize = dataSurface->GetSize();
-        format = dataSurface->GetFormat();
+        customCursor = nsContentUtils::SurfaceToIPCImage(*dataSurface);
       }
     }
   }
 
-  nsDependentCString cursorData(customCursorData ? customCursorData.get() : "",
-                                length);
   if (!mBrowserChild->SendSetCursor(
-          aCursor.mDefaultCursor, hasCustomCursor, cursorData,
-          customCursorSize.width, customCursorSize.height, resolution.mX,
-          resolution.mY, stride, format, aCursor.mHotspotX, aCursor.mHotspotY,
-          force)) {
+          aCursor.mDefaultCursor, std::move(customCursor), resolution.mX,
+          resolution.mY, aCursor.mHotspotX, aCursor.mHotspotY, force)) {
     return;
   }
   mCursor = aCursor;
-}
-
-void PuppetWidget::SetChild(PuppetWidget* aChild) {
-  MOZ_ASSERT(this != aChild, "can't parent a widget to itself");
-  MOZ_ASSERT(!aChild->mChild,
-             "fake widget 'hierarchy' only expected to have one level");
-
-  mChild = aChild;
+  mUpdateCursor = false;
 }
 
 NS_IMETHODIMP
@@ -1008,30 +931,18 @@ void PuppetWidget::PaintNowIfNeeded() {
 
 void PuppetWidget::OnMemoryPressure(layers::MemoryPressureReason aWhy) {
   if (aWhy != MemoryPressureReason::LOW_MEMORY_ONGOING && !mVisible &&
-      mLayerManager && XRE_IsContentProcess()) {
-    mLayerManager->ClearCachedResources();
+      mWindowRenderer && mWindowRenderer->AsWebRender() &&
+      XRE_IsContentProcess()) {
+    mWindowRenderer->AsWebRender()->ClearCachedResources();
   }
 }
 
 bool PuppetWidget::NeedsPaint() {
   // e10s popups are handled by the parent process, so never should be painted
   // here
-  if (XRE_IsContentProcess() &&
-      StaticPrefs::browser_tabs_remote_desktopbehavior() &&
-      mWindowType == eWindowType_popup) {
-    NS_WARNING("Trying to paint an e10s popup in the child process!");
-    return false;
-  }
-
   return mVisible;
 }
 
-float PuppetWidget::GetDPI() { return mDPI; }
-
-double PuppetWidget::GetDefaultScaleInternal() { return mDefaultScale; }
-
-int32_t PuppetWidget::RoundsWidgetCoordinatesTo() { return mRounding; }
-
 LayoutDeviceIntPoint PuppetWidget::GetChromeOffset() {
   if (!GetOwningBrowserChild()) {
     NS_WARNING("PuppetWidget without Tab does not have chrome information.");
@@ -1041,11 +952,7 @@ LayoutDeviceIntPoint PuppetWidget::GetChromeOffset() {
 }
 
 LayoutDeviceIntPoint PuppetWidget::WidgetToScreenOffset() {
-  auto positionRalativeToWindow =
-      WidgetToTopLevelWidgetTransform().TransformPoint(LayoutDevicePoint());
-
-  return GetWindowPosition() +
-         LayoutDeviceIntPoint::Round(positionRalativeToWindow);
+  return GetWindowPosition() + WidgetToTopLevelWidgetOffset();
 }
 
 LayoutDeviceIntPoint PuppetWidget::GetWindowPosition() {
@@ -1054,9 +961,9 @@ LayoutDeviceIntPoint PuppetWidget::GetWindowPosition() {
   }
 
   int32_t winX, winY, winW, winH;
-  NS_ENSURE_SUCCESS(
-      GetOwningBrowserChild()->GetDimensions(0, &winX, &winY, &winW, &winH),
-      LayoutDeviceIntPoint());
+  NS_ENSURE_SUCCESS(GetOwningBrowserChild()->GetDimensions(
+                        DimensionKind::Outer, &winX, &winY, &winW, &winH),
+                    LayoutDeviceIntPoint());
   return LayoutDeviceIntPoint(winX, winY) +
          GetOwningBrowserChild()->GetClientOffset();
 }
@@ -1074,88 +981,12 @@ void PuppetWidget::StartAsyncScrollbarDrag(
   mBrowserChild->StartScrollbarDrag(aDragMetrics);
 }
 
-PuppetScreen::PuppetScreen(void* nativeScreen) {}
-
-PuppetScreen::~PuppetScreen() = default;
-
-static ScreenConfiguration ScreenConfig() {
-  ScreenConfiguration config;
-  hal::GetCurrentScreenConfiguration(&config);
-  return config;
-}
-
-nsIntSize PuppetWidget::GetScreenDimensions() {
-  nsIntRect r = ScreenConfig().rect();
-  return nsIntSize(r.Width(), r.Height());
-}
-
-NS_IMETHODIMP
-PuppetScreen::GetRect(int32_t* outLeft, int32_t* outTop, int32_t* outWidth,
-                      int32_t* outHeight) {
-  nsIntRect r = ScreenConfig().rect();
-  r.GetRect(outLeft, outTop, outWidth, outHeight);
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-PuppetScreen::GetAvailRect(int32_t* outLeft, int32_t* outTop, int32_t* outWidth,
-                           int32_t* outHeight) {
-  return GetRect(outLeft, outTop, outWidth, outHeight);
-}
-
-NS_IMETHODIMP
-PuppetScreen::GetPixelDepth(int32_t* aPixelDepth) {
-  *aPixelDepth = ScreenConfig().pixelDepth();
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-PuppetScreen::GetColorDepth(int32_t* aColorDepth) {
-  *aColorDepth = ScreenConfig().colorDepth();
-  return NS_OK;
-}
-
-NS_IMPL_ISUPPORTS(PuppetScreenManager, nsIScreenManager)
-
-PuppetScreenManager::PuppetScreenManager() {
-  mOneScreen = new PuppetScreen(nullptr);
-}
-
-PuppetScreenManager::~PuppetScreenManager() = default;
-
-NS_IMETHODIMP
-PuppetScreenManager::GetPrimaryScreen(nsIScreen** outScreen) {
-  NS_IF_ADDREF(*outScreen = mOneScreen.get());
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-PuppetScreenManager::GetTotalScreenPixels(int64_t* aTotalScreenPixels) {
-  MOZ_ASSERT(aTotalScreenPixels);
-  if (mOneScreen) {
-    int32_t x, y, width, height;
-    x = y = width = height = 0;
-    mOneScreen->GetRect(&x, &y, &width, &height);
-    *aTotalScreenPixels = width * height;
-  } else {
-    *aTotalScreenPixels = 0;
-  }
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-PuppetScreenManager::ScreenForRect(int32_t inLeft, int32_t inTop,
-                                   int32_t inWidth, int32_t inHeight,
-                                   nsIScreen** outScreen) {
-  return GetPrimaryScreen(outScreen);
-}
-
-ScreenIntMargin PuppetWidget::GetSafeAreaInsets() const {
+LayoutDeviceIntMargin PuppetWidget::GetSafeAreaInsets() const {
   return mSafeAreaInsets;
 }
 
 void PuppetWidget::UpdateSafeAreaInsets(
-    const ScreenIntMargin& aSafeAreaInsets) {
+    const LayoutDeviceIntMargin& aSafeAreaInsets) {
   mSafeAreaInsets = aSafeAreaInsets;
 }
 
@@ -1188,8 +1019,8 @@ void PuppetWidget::LookUpDictionary(
     return;
   }
 
-  mBrowserChild->SendLookUpDictionary(nsString(aText), aFontRangeArray,
-                                      aIsVertical, aPoint);
+  mBrowserChild->SendLookUpDictionary(aText, aFontRangeArray, aIsVertical,
+                                      aPoint);
 }
 
 bool PuppetWidget::HasPendingInputEvent() {
@@ -1247,8 +1078,6 @@ PuppetWidget::NotifyIME(TextEventDispatcher* aTextEventDispatcher,
     default:
       return NS_ERROR_NOT_IMPLEMENTED;
   }
-
-  return NS_ERROR_NOT_IMPLEMENTED;
 }
 
 NS_IMETHODIMP_(IMENotificationRequests)
@@ -1289,5 +1118,9 @@ nsresult PuppetWidget::GetSystemFont(nsCString& aFontName) {
   return NS_OK;
 }
 
+LayersId PuppetWidget::GetLayersId() const {
+  return mBrowserChild ? mBrowserChild->GetLayersId() : LayersId{0};
+}
+
 }  // namespace widget
 }  // namespace mozilla
