@@ -7,14 +7,8 @@
 #define SystemTimeConverter_h
 
 #include <limits>
+#include <type_traits>
 #include "mozilla/TimeStamp.h"
-#include "mozilla/TypeTraits.h"
-
-// GetCurrentTime is defined in winbase.h as zero argument macro forwarding to
-// GetTickCount().
-#ifdef GetCurrentTime
-#undef GetCurrentTime
-#endif
 
 namespace mozilla {
 
@@ -31,29 +25,31 @@ namespace mozilla {
 // For performance reasons, this class is careful to minimize calls to the
 // native "current time" function (e.g. gdk_x11_server_get_time) since this can
 // be slow.
-template <typename Time>
+template <typename Time, typename TimeStampNowProvider = TimeStamp>
 class SystemTimeConverter {
  public:
   SystemTimeConverter()
       : mReferenceTime(Time(0)),
-        mReferenceTimeStamp()  // Initializes to the null timestamp
-        ,
         mLastBackwardsSkewCheck(Time(0)),
         kTimeRange(std::numeric_limits<Time>::max()),
         kTimeHalfRange(kTimeRange / 2),
         kBackwardsSkewCheckInterval(Time(2000)) {
-    static_assert(!IsSigned<Time>::value, "Expected Time to be unsigned");
+    static_assert(!std::is_signed_v<Time>, "Expected Time to be unsigned");
   }
 
   template <typename CurrentTimeGetter>
   mozilla::TimeStamp GetTimeStampFromSystemTime(
       Time aTime, CurrentTimeGetter& aCurrentTimeGetter) {
+    TimeStamp roughlyNow = TimeStampNowProvider::Now();
+
     // If the reference time is not set, use the current time value to fill
     // it in.
     if (mReferenceTimeStamp.IsNull()) {
+      // This sometimes happens when ::GetMessageTime returns 0 for the first
+      // message on Windows.
+      if (!aTime) return roughlyNow;
       UpdateReferenceTime(aTime, aCurrentTimeGetter);
     }
-    TimeStamp roughlyNow = TimeStamp::Now();
 
     // Check for skew between the source of Time values and TimeStamp values.
     // We do this by comparing two durations (both in ms):
@@ -75,7 +71,8 @@ class SystemTimeConverter {
     // the case when we have a backlog of events and by the time we process
     // them, the time given by the system is comparatively "old".
     //
-    // We call the absolute difference between (i) and (ii), "deltaFromNow".
+    // The IsNewerThanTimestamp function computes the equivalent of |aTime| in
+    // the TimeStamp scale and returns that in |timeAsTimeStamp|.
     //
     // Graphically:
     //
@@ -88,13 +85,13 @@ class SystemTimeConverter {
     //                          |------timeStampDelta-------|
     //
     //                                                  |---|
-    //                                               deltaFromNow
+    //                                         roughlyNow-timeAsTimeStamp
     //
-    Time deltaFromNow;
-    bool newer = IsTimeNewerThanTimestamp(aTime, roughlyNow, &deltaFromNow);
+    TimeStamp timeAsTimeStamp;
+    bool newer = IsTimeNewerThanTimestamp(aTime, roughlyNow, &timeAsTimeStamp);
 
     // Tolerance when detecting clock skew.
-    static const Time kTolerance = 30;
+    static const TimeDuration kTolerance = TimeDuration::FromMilliseconds(30.0);
 
     // Check for forwards skew
     if (newer) {
@@ -108,7 +105,7 @@ class SystemTimeConverter {
       return roughlyNow;
     }
 
-    if (deltaFromNow <= kTolerance) {
+    if (roughlyNow - timeAsTimeStamp <= kTolerance) {
       // If the time between event times and TimeStamp values is within
       // the tolerance then assume we don't have clock skew so we can
       // avoid checking for backwards skew for a while.
@@ -119,7 +116,7 @@ class SystemTimeConverter {
     }
 
     // Finally, calculate the timestamp
-    return roughlyNow - TimeDuration::FromMilliseconds(deltaFromNow);
+    return timeAsTimeStamp;
   }
 
   void CompensateForBackwardsSkew(Time aReferenceTime,
@@ -157,8 +154,7 @@ class SystemTimeConverter {
     //
     // If that's not the case, then we probably just got caught behind
     // temporarily.
-    Time delta;
-    if (IsTimeNewerThanTimestamp(aReferenceTime, aLowerBound, &delta)) {
+    if (IsTimeNewerThanTimestamp(aReferenceTime, aLowerBound, nullptr)) {
       return;
     }
 
@@ -178,7 +174,7 @@ class SystemTimeConverter {
   void UpdateReferenceTime(Time aReferenceTime,
                            const CurrentTimeGetter& aCurrentTimeGetter) {
     Time currentTime = aCurrentTimeGetter.GetCurrentTime();
-    TimeStamp currentTimeStamp = TimeStamp::Now();
+    TimeStamp currentTimeStamp = TimeStampNowProvider::Now();
     Time timeSinceReference = currentTime - aReferenceTime;
     TimeStamp referenceTimeStamp =
         currentTimeStamp - TimeDuration::FromMilliseconds(timeSinceReference);
@@ -192,7 +188,7 @@ class SystemTimeConverter {
   }
 
   bool IsTimeNewerThanTimestamp(Time aTime, TimeStamp aTimeStamp,
-                                Time* aDelta) {
+                                TimeStamp* aTimeAsTimeStamp) {
     Time timeDelta = aTime - mReferenceTime;
 
     // Cast the result to signed 64-bit integer first since that should be
@@ -201,18 +197,41 @@ class SystemTimeConverter {
     // is outside the integer range is undefined.
     // Then we do an implicit cast to Time (typically an unsigned 32-bit
     // integer) which wraps times outside that range.
-    Time timeStampDelta = static_cast<int64_t>(
-        (aTimeStamp - mReferenceTimeStamp).ToMilliseconds());
+    TimeDuration timeStampDelta = (aTimeStamp - mReferenceTimeStamp);
+    int64_t wholeMillis = static_cast<int64_t>(timeStampDelta.ToMilliseconds());
+    Time wrappedTimeStampDelta = wholeMillis;  // truncate to unsigned
+    // Half of the valid range of Time
+    const Time shift = (static_cast<Time>(0) - static_cast<Time>(1)) / 2;
+    // Shift/move origin (0) of the value by UINT32_MAX / 2 and shift
+    // it back later in order to support negative deltas. With this
+    // approach we can support deltas before shifting in the range
+    // [UINT32_MAX / 2 + 1, -UINT32_MAX / 2].
+    Time wrappedTimeStampDeltaShifted = wrappedTimeStampDelta + shift;
 
-    Time timeToTimeStamp = timeStampDelta - timeDelta;
+    int64_t timeToTimeStamp = static_cast<int64_t>(wrappedTimeStampDeltaShifted) -
+                              static_cast<int64_t>(timeDelta) -
+                              static_cast<int64_t>(shift);
     bool isNewer = false;
     if (timeToTimeStamp == 0) {
-      *aDelta = 0;
-    } else if (timeToTimeStamp < kTimeHalfRange) {
-      *aDelta = timeToTimeStamp;
-    } else {
+      // wholeMillis needs no adjustment
+    } else if (timeToTimeStamp < 0) {
       isNewer = true;
-      *aDelta = timeDelta - timeStampDelta;
+      wholeMillis += (-timeToTimeStamp);
+    } else {
+      wholeMillis -= timeToTimeStamp;
+    }
+    if (aTimeAsTimeStamp) {
+      *aTimeAsTimeStamp =
+          mReferenceTimeStamp + TimeDuration::FromMilliseconds(wholeMillis);
+
+      if (aTimeAsTimeStamp->IsNull()) {
+        MOZ_CRASH_UNSAFE_PRINTF(
+            "Failed to compute the new timestamp, aTime: %" PRIu32
+            ", timeDelta: %" PRIu32 ", wholeMillis: %" PRId64
+            ", timeToTimeStamp: %" PRId64,
+            static_cast<uint32_t>(aTime), static_cast<uint32_t>(timeDelta),
+            wholeMillis, timeToTimeStamp);
+      }
     }
 
     return isNewer;
