@@ -1,4 +1,4 @@
-/* -*- Mode: c++; c-basic-offset: 4; tab-width: 20; indent-tabs-mode: nil; -*-
+/* -*- Mode: c++; c-basic-offset: 2; tab-width: 20; indent-tabs-mode: nil; -*-
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
@@ -6,11 +6,12 @@
 #ifndef GeckoNetworkManager_h
 #define GeckoNetworkManager_h
 
-#include "GeneratedJNINatives.h"
 #include "nsAppShell.h"
 #include "nsCOMPtr.h"
 #include "nsINetworkLinkService.h"
+#include "nsISystemProxySettings.h"
 
+#include "mozilla/java/GeckoNetworkManagerNatives.h"
 #include "mozilla/Services.h"
 
 namespace mozilla {
@@ -38,6 +39,27 @@ class GeckoNetworkManager final
                           aStatus->ToString().get());
     }
   }
+
+  static void OnProxyChanged(jni::String::Param aHost, int32_t aPort,
+                             jni::String::Param aPacFileUrl,
+                             jni::ObjectArray::Param aExclusionList) {
+    nsCOMPtr<nsISystemProxySettings> sp =
+        do_GetService(NS_SYSTEMPROXYSETTINGS_CONTRACTID);
+    if (!sp) {
+      return;
+    }
+    nsAutoCString host = NS_ConvertUTF16toUTF8(aHost->ToString().get());
+    nsAutoCString pacFileUrl =
+        NS_ConvertUTF16toUTF8(aPacFileUrl->ToString().get());
+    jni::ObjectArray::LocalRef exclusions = aExclusionList;
+    nsTArray<nsCString> exclusionList;
+    for (size_t i = 0; i < exclusions->Length(); i++) {
+      jni::String::LocalRef exclusion = exclusions->GetElement(i);
+      exclusionList.AppendElement(exclusion->ToCString());
+    }
+
+    sp->SetSystemProxyInfo(host, aPort, pacFileUrl, exclusionList);
+  }
 };
 
 }  // namespace mozilla
