@@ -195,9 +195,6 @@ void GfxInfo::EnsureInitialized() {
       ", Manufacturer: %s", NS_LossyConvertUTF16toASCII(mManufacturer).get());
 
   mSDKVersion = java::sdk::Build::VERSION::SDK_INT();
-  // the HARDWARE field isn't available on Android SDK < 8, but we require 9+
-  // anyway.
-  MOZ_ASSERT(mSDKVersion >= 8);
   jni::String::LocalRef hardware = java::sdk::Build::HARDWARE();
   mHardware = hardware->ToString();
   mAdapterDescription.AppendPrintf(
@@ -217,6 +214,8 @@ void GfxInfo::EnsureInitialized() {
   mOSVersionInteger = (uint32_t(na) << 24) | (uint32_t(nb) << 16) |
                       (uint32_t(nc) << 8) | uint32_t(nd);
 
+  mOSVersionEx.Parse(mOSVersion);
+
   mAdapterDescription.AppendPrintf(
       ", OpenGL: %s -- %s -- %s", mGLStrings->Vendor().get(),
       mGLStrings->Renderer().get(), mGLStrings->Version().get());
@@ -353,15 +352,15 @@ GfxInfo::GetDrmRenderDevice(nsACString& aDrmRenderDevice) {
 }
 
 void GfxInfo::AddCrashReportAnnotations() {
-  CrashReporter::AnnotateCrashReport(CrashReporter::Annotation::AdapterVendorID,
-                                     mGLStrings->Vendor());
-  CrashReporter::AnnotateCrashReport(CrashReporter::Annotation::AdapterDeviceID,
-                                     mGLStrings->Renderer());
-  CrashReporter::AnnotateCrashReport(
+  CrashReporter::RecordAnnotationNSCString(
+      CrashReporter::Annotation::AdapterVendorID, mGLStrings->Vendor());
+  CrashReporter::RecordAnnotationNSCString(
+      CrashReporter::Annotation::AdapterDeviceID, mGLStrings->Renderer());
+  CrashReporter::RecordAnnotationNSCString(
       CrashReporter::Annotation::AdapterDriverVersion, mGLStrings->Version());
 }
 
-const nsTArray<GfxDriverInfo>& GfxInfo::GetGfxDriverInfo() {
+const nsTArray<RefPtr<GfxDriverInfo>>& GfxInfo::GetGfxDriverInfo() {
   if (sDriverInfo->IsEmpty()) {
     APPEND_TO_DRIVER_BLOCKLIST2(
         OperatingSystem::Android, DeviceFamily::All,
@@ -375,7 +374,7 @@ const nsTArray<GfxDriverInfo>& GfxInfo::GetGfxDriverInfo() {
 
 nsresult GfxInfo::GetFeatureStatusImpl(
     int32_t aFeature, int32_t* aStatus, nsAString& aSuggestedDriverVersion,
-    const nsTArray<GfxDriverInfo>& aDriverInfo, nsACString& aFailureId,
+    const nsTArray<RefPtr<GfxDriverInfo>>& aDriverInfo, nsACString& aFailureId,
     OperatingSystem* aOS /* = nullptr */) {
   NS_ENSURE_ARG_POINTER(aStatus);
   aSuggestedDriverVersion.SetIsVoid(true);
@@ -410,11 +409,7 @@ nsresult GfxInfo::GetFeatureStatusImpl(
   // Don't evaluate special cases when evaluating the downloaded blocklist.
   if (aDriverInfo.IsEmpty()) {
     if (aFeature == nsIGfxInfo::FEATURE_CANVAS2D_ACCELERATION) {
-      if (mSDKVersion < 11) {
-        // It's slower than software due to not having a compositing fast path
-        *aStatus = nsIGfxInfo::FEATURE_BLOCKED_OS_VERSION;
-        aFailureId = "FEATURE_FAILURE_CANVAS_2D_SDK";
-      } else if (mGLStrings->Renderer().Find("Vivante GC1000") != -1) {
+      if (mGLStrings->Renderer().Find("Vivante GC1000") != -1) {
         // Blocklist Vivante GC1000. See bug 1248183.
         *aStatus = nsIGfxInfo::FEATURE_BLOCKED_DEVICE;
         aFailureId = "FEATURE_FAILED_CANVAS_2D_HW";
@@ -432,14 +427,6 @@ nsresult GfxInfo::GetFeatureStatusImpl(
         return NS_OK;
       }
 
-      if (mSDKVersion <= 17) {
-        if (mGLStrings->Renderer().Find("Adreno (TM) 3") != -1) {
-          *aStatus = nsIGfxInfo::FEATURE_BLOCKED_DEVICE;
-          aFailureId = "FEATURE_FAILURE_ADRENO_3xx";
-        }
-        return NS_OK;
-      }
-
       if (mHardware.EqualsLiteral("ville")) {
         *aStatus = nsIGfxInfo::FEATURE_BLOCKED_DEVICE;
         aFailureId = "FEATURE_FAILURE_VILLE";
@@ -682,6 +669,26 @@ nsresult GfxInfo::GetFeatureStatusImpl(
     return NS_OK;
   }
 
+  if (aFeature == FEATURE_WEBGPU) {
+    // Ensure WebGPU is disabled by default on Android until it is better
+    // tested.
+    *aStatus = nsIGfxInfo::FEATURE_BLOCKED_DEVICE;
+    aFailureId = "FEATURE_FAILURE_WEBGPU_ANDROID";
+    return NS_OK;
+  }
+
+  if (aFeature == FEATURE_GL_NORM16_TEXTURES) {
+    const auto& extensions = mGLStrings->Extensions();
+    // GL_EXT_texture_norm16 is required for GL_R16, GL_RG16, etc texture
+    // formats on OpenGL ES.
+    if (!extensions.Contains("GL_EXT_texture_norm16")) {
+      *aStatus = nsIGfxInfo::FEATURE_BLOCKED_DEVICE;
+      aFailureId = "FEATURE_FAILURE_MISSING_EXTENSION";
+    } else {
+      *aStatus = nsIGfxInfo::FEATURE_STATUS_OK;
+    }
+  }
+
   return GfxInfoBase::GetFeatureStatusImpl(
       aFeature, aStatus, aSuggestedDriverVersion, aDriverInfo, aFailureId, &os);
 }
@@ -840,6 +847,13 @@ NS_IMETHODIMP GfxInfo::SpoofOSVersion(uint32_t aVersion) {
   return NS_OK;
 }
 
+NS_IMETHODIMP GfxInfo::SpoofOSVersionEx(uint32_t aMajor, uint32_t aMinor,
+                                        uint32_t aBuild, uint32_t aRevision) {
+  EnsureInitialized();
+  mOSVersionEx = GfxVersionEx(aMajor, aMinor, aBuild, aRevision);
+  return NS_OK;
+}
+
 #endif
 
 nsString GfxInfo::Model() {
@@ -867,5 +881,10 @@ uint32_t GfxInfo::OperatingSystemVersion() {
   return mOSVersionInteger;
 }
 
+GfxVersionEx GfxInfo::OperatingSystemVersionEx() {
+  EnsureInitialized();
+  return mOSVersionEx;
+}
+
 }  // namespace widget
 }  // namespace mozilla
