@@ -6,8 +6,7 @@
 #include "GtkCompositorWidget.h"
 
 #include "mozilla/gfx/gfxVars.h"
-#include "mozilla/layers/CompositorThread.h"
-#include "mozilla/widget/InProcessCompositorWidget.h"
+#include "mozilla/WidgetUtilsGtk.h"
 #include "mozilla/widget/PlatformWidgetTypes.h"
 #include "nsWindow.h"
 
@@ -39,26 +38,30 @@ GtkCompositorWidget::GtkCompositorWidget(
                   "GtkCompositorWidget::mClientSize") {
 #if defined(MOZ_X11)
   if (GdkIsX11Display()) {
-    mXWindow = (Window)aInitData.XWindow();
-    ConfigureX11Backend(mXWindow, aInitData.Shaped());
-    LOG("GtkCompositorWidget::GtkCompositorWidget() [%p] mXWindow %p "
-        "mIsRenderingSuspended %d\n",
-        (void*)mWidget.get(), (void*)mXWindow, !!mIsRenderingSuspended);
+    ConfigureX11Backend((Window)aInitData.XWindow());
+    LOG("GtkCompositorWidget::GtkCompositorWidget() [%p] mXWindow %p\n",
+        (void*)mWidget.get(), (void*)aInitData.XWindow());
   }
 #endif
 #if defined(MOZ_WAYLAND)
   if (GdkIsWaylandDisplay()) {
     ConfigureWaylandBackend();
-    LOG("GtkCompositorWidget::GtkCompositorWidget() [%p] mWidget %p "
-        "mIsRenderingSuspended %d\n",
-        (void*)mWidget.get(), (void*)mWidget, !!mIsRenderingSuspended);
+    LOG("GtkCompositorWidget::GtkCompositorWidget() [%p] mWidget %p\n",
+        (void*)mWidget.get(), (void*)mWidget);
   }
 #endif
 }
 
 GtkCompositorWidget::~GtkCompositorWidget() {
   LOG("GtkCompositorWidget::~GtkCompositorWidget [%p]\n", (void*)mWidget.get());
-  DisableRendering();
+  CleanupResources();
+#ifdef MOZ_WAYLAND
+  if (mNativeLayerRoot) {
+    NS_DispatchToMainThread(NS_NewRunnableFunction(
+        "~GtkCompositorWidget::NativeLayerRootWayland::Shutdown()",
+        [root = RefPtr{mNativeLayerRoot}]() -> void { root->Shutdown(); }));
+  }
+#endif
   RefPtr<nsIWidget> widget = mWidget.forget();
   NS_ReleaseOnMainThread("GtkCompositorWidget::mWidget", widget.forget());
 }
@@ -70,9 +73,8 @@ void GtkCompositorWidget::EndRemoteDrawing() {}
 
 already_AddRefed<gfx::DrawTarget>
 GtkCompositorWidget::StartRemoteDrawingInRegion(
-    const LayoutDeviceIntRegion& aInvalidRegion,
-    layers::BufferMode* aBufferMode) {
-  return mProvider.StartRemoteDrawingInRegion(aInvalidRegion, aBufferMode);
+    const LayoutDeviceIntRegion& aInvalidRegion) {
+  return mProvider.StartRemoteDrawingInRegion(aInvalidRegion);
 }
 
 void GtkCompositorWidget::EndRemoteDrawingInRegion(
@@ -96,35 +98,14 @@ LayoutDeviceIntSize GtkCompositorWidget::GetClientSize() {
   return *size;
 }
 
-void GtkCompositorWidget::RemoteLayoutSizeUpdated(
-    const LayoutDeviceRect& aSize) {
-  if (!mWidget || !mWidget->IsWaitingForCompositorResume()) {
-    return;
-  }
-
-  LOG("GtkCompositorWidget::RemoteLayoutSizeUpdated() %d x %d",
-      (int)aSize.width, (int)aSize.height);
-
-  // We're waiting for layout to match widget size.
-  auto clientSize = mClientSize.Lock();
-  if (clientSize->width != (int)aSize.width ||
-      clientSize->height != (int)aSize.height) {
-    LOG("quit, client size doesn't match (%d x %d)", clientSize->width,
-        clientSize->height);
-    return;
-  }
-
-  mWidget->ResumeCompositorFromCompositorThread();
-}
-
 EGLNativeWindowType GtkCompositorWidget::GetEGLNativeWindow() {
   EGLNativeWindowType window = nullptr;
   if (mWidget) {
     window = (EGLNativeWindowType)mWidget->GetNativeData(NS_NATIVE_EGL_WINDOW);
   }
 #if defined(MOZ_X11)
-  if (mXWindow) {
-    window = (EGLNativeWindowType)mXWindow;
+  else {
+    window = (EGLNativeWindowType)mProvider.GetXWindow();
   }
 #endif
   LOG("GtkCompositorWidget::GetEGLNativeWindow [%p] window %p\n", mWidget.get(),
@@ -132,25 +113,24 @@ EGLNativeWindowType GtkCompositorWidget::GetEGLNativeWindow() {
   return window;
 }
 
-#if defined(MOZ_WAYLAND)
-void GtkCompositorWidget::SetEGLNativeWindowSize(
+bool GtkCompositorWidget::SetEGLNativeWindowSize(
     const LayoutDeviceIntSize& aEGLWindowSize) {
-  if (mWidget) {
-    mWidget->SetEGLNativeWindowSize(aEGLWindowSize);
+#if defined(MOZ_WAYLAND)
+  // We explicitly need to set EGL window size on Wayland only.
+  if (GdkIsWaylandDisplay() && mWidget) {
+    return mWidget->SetEGLNativeWindowSize(aEGLWindowSize);
   }
-}
 #endif
+  return true;
+}
 
 LayoutDeviceIntRegion GtkCompositorWidget::GetTransparentRegion() {
-  // We need to clear target buffer alpha values of popup windows as
-  // SW-WR paints with alpha blending (see Bug 1674473).
-  if (!mWidget || mWidget->IsPopup()) {
-    return LayoutDeviceIntRect(LayoutDeviceIntPoint(0, 0), GetClientSize());
+  LayoutDeviceIntRegion fullRegion(
+      LayoutDeviceIntRect(LayoutDeviceIntPoint(), GetClientSize()));
+  if (mWidget) {
+    fullRegion.SubOut(mWidget->GetOpaqueRegion());
   }
-
-  // Clear background of titlebar area to render titlebar
-  // transparent corners correctly.
-  return mWidget->GetTitlebarRect();
+  return fullRegion;
 }
 
 #ifdef MOZ_WAYLAND
@@ -158,9 +138,12 @@ RefPtr<mozilla::layers::NativeLayerRoot>
 GtkCompositorWidget::GetNativeLayerRoot() {
   if (gfx::gfxVars::UseWebRenderCompositor()) {
     if (!mNativeLayerRoot) {
+      LOG("GtkCompositorWidget::GetNativeLayerRoot [%p] create",
+          (void*)mWidget.get());
       MOZ_ASSERT(mWidget && mWidget->GetMozContainer());
-      mNativeLayerRoot = NativeLayerRootWayland::CreateForMozContainer(
-          mWidget->GetMozContainer());
+      mNativeLayerRoot = layers::NativeLayerRootWayland::Create(
+          MOZ_WL_SURFACE(mWidget->GetMozContainer()));
+      mNativeLayerRoot->Init();
     }
     return mNativeLayerRoot;
   }
@@ -168,79 +151,55 @@ GtkCompositorWidget::GetNativeLayerRoot() {
 }
 #endif
 
-void GtkCompositorWidget::DisableRendering() {
-  LOG("GtkCompositorWidget::DisableRendering [%p]\n", (void*)mWidget.get());
-  mIsRenderingSuspended = true;
+void GtkCompositorWidget::CleanupResources() {
+  LOG("GtkCompositorWidget::CleanupResources [%p]\n", (void*)mWidget.get());
   mProvider.CleanupResources();
-#if defined(MOZ_X11)
-  mXWindow = {};
-#endif
 }
 
 #if defined(MOZ_WAYLAND)
-bool GtkCompositorWidget::ConfigureWaylandBackend() {
+void GtkCompositorWidget::ConfigureWaylandBackend() {
   mProvider.Initialize(this);
-  return true;
 }
 #endif
 
 #if defined(MOZ_X11)
-bool GtkCompositorWidget::ConfigureX11Backend(Window aXWindow, bool aShaped) {
-  mXWindow = aXWindow;
-
+void GtkCompositorWidget::ConfigureX11Backend(Window aXWindow) {
   // We don't have X window yet.
-  if (!mXWindow) {
-    mIsRenderingSuspended = true;
-    return false;
+  if (!aXWindow) {
+    mProvider.CleanupResources();
+    return;
   }
-
-  // Grab the window's visual and depth
-  XWindowAttributes windowAttrs;
-  if (!XGetWindowAttributes(DefaultXDisplay(), mXWindow, &windowAttrs)) {
-    NS_WARNING("GtkCompositorWidget(): XGetWindowAttributes() failed!");
-    return false;
-  }
-
-  Visual* visual = windowAttrs.visual;
-  int depth = windowAttrs.depth;
-
   // Initialize the window surface provider
-  mProvider.Initialize(mXWindow, visual, depth, aShaped);
-  return true;
+  mProvider.Initialize(aXWindow);
 }
 #endif
 
-void GtkCompositorWidget::EnableRendering(const uintptr_t aXWindow,
-                                          const bool aShaped) {
-  LOG("GtkCompositorWidget::EnableRendering() [%p]\n", mWidget.get());
+void GtkCompositorWidget::SetRenderingSurface(const uintptr_t aXWindow) {
+  LOG("GtkCompositorWidget::SetRenderingSurface() [%p]\n", mWidget.get());
 
-  if (!mIsRenderingSuspended) {
-    LOG("  quit, mIsRenderingSuspended = false\n");
-    return;
-  }
 #if defined(MOZ_WAYLAND)
   if (GdkIsWaylandDisplay()) {
     LOG("  configure widget %p\n", mWidget.get());
-    if (!ConfigureWaylandBackend()) {
-      return;
-    }
+    ConfigureWaylandBackend();
   }
 #endif
 #if defined(MOZ_X11)
   if (GdkIsX11Display()) {
-    LOG("  configure XWindow %p shaped %d\n", (void*)aXWindow, aShaped);
-    if (!ConfigureX11Backend((Window)aXWindow, aShaped)) {
-      return;
-    }
+    LOG("  configure XWindow %p\n", (void*)aXWindow);
+    ConfigureX11Backend((Window)aXWindow);
   }
 #endif
-  mIsRenderingSuspended = false;
 }
+
 #ifdef MOZ_LOGGING
 bool GtkCompositorWidget::IsPopup() {
   return mWidget ? mWidget->IsPopup() : false;
 }
 #endif
 
+UniquePtr<WaylandSurfaceLock> GtkCompositorWidget::LockSurface() {
+  return mWidget ? mWidget->LockSurface() : nullptr;
+}
+
 }  // namespace widget
 }  // namespace mozilla
