@@ -13,12 +13,15 @@
 
 #include "CompositorWidget.h"
 #include "MozContainer.h"
+#include "WaylandSurfaceLock.h"
 #include "VsyncSource.h"
 #include "mozilla/EventForwards.h"
 #include "mozilla/Maybe.h"
 #include "mozilla/RefPtr.h"
 #include "mozilla/TouchEvents.h"
 #include "mozilla/UniquePtr.h"
+#include "mozilla/RWLock.h"
+#include "mozilla/gfx/BaseMargin.h"
 #include "mozilla/widget/WindowSurface.h"
 #include "mozilla/widget/WindowSurfaceProvider.h"
 #include "nsBaseWidget.h"
@@ -26,6 +29,7 @@
 #include "nsIDragService.h"
 #include "nsRefPtrHashtable.h"
 #include "IMContextWrapper.h"
+#include "LookAndFeel.h"
 
 #ifdef ACCESSIBILITY
 #  include "mozilla/a11y/LocalAccessible.h"
@@ -38,12 +42,14 @@
 #ifdef MOZ_WAYLAND
 #  include <gdk/gdkwayland.h>
 #  include "base/thread.h"
-#  include "WaylandVsyncSource.h"
 #  include "nsClipboardWayland.h"
 #endif
 
 #ifdef MOZ_LOGGING
 
+#  undef LOG
+#  undef LOGVERBOSE
+
 #  include "mozilla/Logging.h"
 #  include "nsTArray.h"
 #  include "Units.h"
@@ -52,25 +58,37 @@ extern mozilla::LazyLogModule gWidgetLog;
 extern mozilla::LazyLogModule gWidgetDragLog;
 extern mozilla::LazyLogModule gWidgetPopupLog;
 extern mozilla::LazyLogModule gWidgetVsync;
+extern mozilla::LazyLogModule gWidgetWaylandLog;
 
 #  define LOG(str, ...)                               \
     MOZ_LOG(IsPopup() ? gWidgetPopupLog : gWidgetLog, \
             mozilla::LogLevel::Debug,                 \
             ("%s: " str, GetDebugTag().get(), ##__VA_ARGS__))
+#  define LOGVERBOSE(str, ...)                        \
+    MOZ_LOG(IsPopup() ? gWidgetPopupLog : gWidgetLog, \
+            mozilla::LogLevel::Verbose,               \
+            ("%s: " str, GetDebugTag().get(), ##__VA_ARGS__))
 #  define LOGW(...) MOZ_LOG(gWidgetLog, mozilla::LogLevel::Debug, (__VA_ARGS__))
 #  define LOGDRAG(...) \
     MOZ_LOG(gWidgetDragLog, mozilla::LogLevel::Debug, (__VA_ARGS__))
 #  define LOG_POPUP(...) \
     MOZ_LOG(gWidgetPopupLog, mozilla::LogLevel::Debug, (__VA_ARGS__))
-#  define LOG_VSYNC(...) \
-    MOZ_LOG(gWidgetVsync, mozilla::LogLevel::Debug, (__VA_ARGS__))
+#  define LOG_VSYNC(str, ...)                       \
+    MOZ_LOG(gWidgetVsync, mozilla::LogLevel::Debug, \
+            ("%s: " str, GetDebugTag().get(), ##__VA_ARGS__))
 #  define LOG_ENABLED()                                         \
     (MOZ_LOG_TEST(gWidgetPopupLog, mozilla::LogLevel::Debug) || \
      MOZ_LOG_TEST(gWidgetLog, mozilla::LogLevel::Debug))
+#  define LOG_ENABLED_VERBOSE()                                   \
+    (MOZ_LOG_TEST(gWidgetPopupLog, mozilla::LogLevel::Verbose) || \
+     MOZ_LOG_TEST(gWidgetLog, mozilla::LogLevel::Verbose))
+#  define LOG_WAYLAND(...) \
+    MOZ_LOG(gWidgetWaylandLog, mozilla::LogLevel::Debug, (__VA_ARGS__))
 
 #else
 
 #  define LOG(...)
+#  define LOGVERBOSE(...)
 #  define LOGW(...)
 #  define LOGDRAG(...)
 #  define LOG_POPUP(...)
@@ -103,19 +121,37 @@ typedef enum {
 } GdkAnchorHints;
 #endif
 
+#if !GTK_CHECK_VERSION(3, 18, 0)
+typedef enum {
+  GDK_TOUCHPAD_GESTURE_PHASE_BEGIN,
+  GDK_TOUCHPAD_GESTURE_PHASE_UPDATE,
+  GDK_TOUCHPAD_GESTURE_PHASE_END,
+  GDK_TOUCHPAD_GESTURE_PHASE_CANCEL
+} GdkTouchpadGesturePhase;
+#endif
+
+struct zwp_locked_pointer_v1;
+struct zwp_relative_pointer_v1;
+
 namespace mozilla {
 enum class NativeKeyBindingsType : uint8_t;
 
 class TimeStamp;
+class WaylandVsyncSource;
 #ifdef MOZ_X11
 class CurrentX11TimeGetter;
 #endif
 
 namespace widget {
+class DBusMenuBar;
 class Screen;
+class WaylandSurface;
+class WaylandSurfaceLock;
 }  // namespace widget
 }  // namespace mozilla
 
+class gfxImageSurface;
+
 class nsWindow final : public nsBaseWidget {
  public:
   typedef mozilla::gfx::DrawTarget DrawTarget;
@@ -137,27 +173,24 @@ class nsWindow final : public nsBaseWidget {
   void OnDestroy() override;
 
   // called to check and see if a widget's dimensions are sane
-  bool AreBoundsSane(void);
+  bool AreBoundsSane();
 
   // nsIWidget
   using nsBaseWidget::Create;  // for Create signature not overridden here
   [[nodiscard]] nsresult Create(nsIWidget* aParent,
-                                nsNativeWidget aNativeParent,
                                 const LayoutDeviceIntRect& aRect,
                                 InitData* aInitData) override;
   void Destroy() override;
-  nsIWidget* GetParent() override;
   float GetDPI() override;
   double GetDefaultScaleInternal() override;
   mozilla::DesktopToLayoutDeviceScale GetDesktopToDeviceScale() override;
   mozilla::DesktopToLayoutDeviceScale GetDesktopToDeviceScaleByScreen()
       override;
-  void SetParent(nsIWidget* aNewParent) override;
   void SetModal(bool aModal) override;
   bool IsVisible() const override;
   bool IsMapped() const override;
   void ConstrainPosition(DesktopIntPoint&) override;
-  void SetSizeConstraints(const SizeConstraints& aConstraints) override;
+  void SetSizeConstraints(const SizeConstraints&) override;
   void LockAspectRatio(bool aShouldLock) override;
   void Move(double aX, double aY) override;
   void Show(bool aState) override;
@@ -166,7 +199,6 @@ class nsWindow final : public nsBaseWidget {
               bool aRepaint) override;
   bool IsEnabled() const override;
 
-  void SetZIndex(int32_t aZIndex) override;
   nsSizeMode SizeMode() override { return mSizeMode; }
   void SetSizeMode(nsSizeMode aMode) override;
   void GetWorkspaceID(nsAString& workspaceID) override;
@@ -176,17 +208,23 @@ class nsWindow final : public nsBaseWidget {
   LayoutDeviceIntRect GetScreenBounds() override;
   LayoutDeviceIntRect GetClientBounds() override;
   LayoutDeviceIntSize GetClientSize() override;
-  LayoutDeviceIntPoint GetClientOffset() override { return mClientOffset; }
+  LayoutDeviceIntPoint GetClientOffset() override {
+    return LayoutDeviceIntPoint(mClientMargin.left, mClientMargin.top);
+  }
+  GdkPoint GetCsdOffsetInGdkCoords();
   LayoutDeviceIntPoint GetScreenEdgeSlop() override;
+  nsresult GetRestoredBounds(LayoutDeviceIntRect&) override;
+  bool PersistClientBounds() const override { return true; }
+  LayoutDeviceIntMargin NormalSizeModeClientToWindowMargin() override;
 
-  // Recomputes the client offset according to our current window position.
-  // If aNotify is true, NotifyWindowMoved will be called on client offset
-  // changes.
-  //
-  // NOTE(emilio): It seems that as long any change here update either the size
-  // or the position of the window, we should be doing fine without notifying,
-  // but this is done to preserve existing behavior.
-  void RecomputeClientOffset(bool aNotify);
+  void ConstrainSize(int* aWidth, int* aHeight) override;
+
+  // Recomputes the bounds according to our current window position. Dispatches
+  // move / resizes as needed.
+  enum class MayChangeCsdMargin : bool { No = false, Yes };
+  void RecomputeBounds(MayChangeCsdMargin);
+  void SchedulePendingBounds(MayChangeCsdMargin);
+  void MaybeRecomputeBounds();
 
   void SetCursor(const Cursor&) override;
   void Invalidate(const LayoutDeviceIntRect& aRect) override;
@@ -225,11 +263,10 @@ class nsWindow final : public nsBaseWidget {
 
   // event callbacks
   gboolean OnExposeEvent(cairo_t* cr);
-  gboolean OnConfigureEvent(GtkWidget* aWidget, GdkEventConfigure* aEvent);
+  gboolean OnShellConfigureEvent(GdkEventConfigure* aEvent);
+  void OnContainerSizeAllocate(GtkAllocation* aAllocation);
   void OnMap();
   void OnUnmap();
-  void OnUnrealize();
-  void OnSizeAllocate(GtkAllocation* aAllocation);
   void OnDeleteEvent();
   void OnEnterNotifyEvent(GdkEventCrossing* aEvent);
   void OnLeaveNotifyEvent(GdkEventCrossing* aEvent);
@@ -242,7 +279,9 @@ class nsWindow final : public nsBaseWidget {
   gboolean OnKeyReleaseEvent(GdkEventKey* aEvent);
 
   void OnScrollEvent(GdkEventScroll* aEvent);
+  void OnSmoothScrollEvent(uint32_t aTime, float aDeltaX, float aDeltaY);
 
+  void OnVisibilityNotifyEvent(GdkVisibilityState aState);
   void OnWindowStateEvent(GtkWidget* aWidget, GdkEventWindowState* aEvent);
   void OnDragDataReceivedEvent(GtkWidget* aWidget, GdkDragContext* aDragContext,
                                gint aX, gint aY,
@@ -251,14 +290,20 @@ class nsWindow final : public nsBaseWidget {
   gboolean OnPropertyNotifyEvent(GtkWidget* aWidget, GdkEventProperty* aEvent);
   gboolean OnTouchEvent(GdkEventTouch* aEvent);
   gboolean OnTouchpadPinchEvent(GdkEventTouchpadPinch* aEvent);
+  void OnTouchpadHoldEvent(GdkTouchpadGesturePhase aPhase, guint aTime,
+                           uint32_t aFingers);
 
   gint GetInputRegionMarginInGdkCoords();
 
-  void UpdateTopLevelOpaqueRegion();
+  void UpdateOpaqueRegionInternal();
+  void UpdateOpaqueRegion(const LayoutDeviceIntRegion&) override;
+  LayoutDeviceIntRegion GetOpaqueRegionForTesting() const override {
+    return GetOpaqueRegion();
+  }
+  LayoutDeviceIntRegion GetOpaqueRegion() const;
 
   already_AddRefed<mozilla::gfx::DrawTarget> StartRemoteDrawingInRegion(
-      const LayoutDeviceIntRegion& aInvalidRegion,
-      mozilla::layers::BufferMode* aBufferMode) override;
+      const LayoutDeviceIntRegion& aInvalidRegion) override;
   void EndRemoteDrawingInRegion(
       mozilla::gfx::DrawTarget* aDrawTarget,
       const LayoutDeviceIntRegion& aInvalidRegion) override;
@@ -268,11 +313,19 @@ class nsWindow final : public nsBaseWidget {
   RefPtr<mozilla::VsyncDispatcher> GetVsyncDispatcher() override;
   bool SynchronouslyRepaintOnResize() override;
 
-  void OnDPIChanged(void);
-  void OnCheckResize(void);
-  void OnCompositedChanged(void);
-  void OnScaleChanged();
+  void OnDPIChanged();
+  void OnCheckResize();
+  void OnCompositedChanged();
   void DispatchResized();
+  void OnScaleEvent();
+
+  // Load new scale from system.
+
+  // aRefreshScreen means notify layout that scale was changed and
+  //  it should load correct values.
+  // Set aRefreshScreen to false if we operate on hidden window
+  // or if we're going to repaint.
+  void RefreshScale(bool aRefreshScreen);
 
   static guint32 sLastButtonPressTime;
 
@@ -285,7 +338,6 @@ class nsWindow final : public nsBaseWidget {
   bool IsDestroyed() const { return mIsDestroyed; }
   bool IsPopup() const;
   bool IsWaylandPopup() const;
-  bool IsPIPWindow() const { return mIsPIPWindow; };
   bool IsDragPopup() { return mIsDragPopup; };
 
   nsAutoCString GetDebugTag() const;
@@ -311,70 +363,58 @@ class nsWindow final : public nsBaseWidget {
       const mozilla::WidgetKeyboardEvent& aEvent,
       nsTArray<mozilla::CommandInt>& aCommands) override;
 
-  // These methods are for toplevel windows only.
-  void ResizeTransparencyBitmap();
-  void ApplyTransparencyBitmap();
-  void ClearTransparencyBitmap();
-
   void SetTransparencyMode(TransparencyMode aMode) override;
   TransparencyMode GetTransparencyMode() override;
   void SetInputRegion(const InputRegion&) override;
-  nsresult UpdateTranslucentWindowAlphaInternal(const nsIntRect& aRect,
-                                                uint8_t* aAlphas,
-                                                int32_t aStride);
-  void ReparentNativeWidget(nsIWidget* aNewParent) override;
 
-  void UpdateTitlebarTransparencyBitmap();
+  nsresult SynthesizeNativeMouseEvent(
+      LayoutDeviceIntPoint aPoint, NativeMouseMessage aNativeMessage,
+      mozilla::MouseButton aButton, nsIWidget::Modifiers aModifierFlags,
+      nsISynthesizedEventCallback* aCallback) override;
 
-  nsresult SynthesizeNativeMouseEvent(LayoutDeviceIntPoint aPoint,
-                                      NativeMouseMessage aNativeMessage,
-                                      mozilla::MouseButton aButton,
-                                      nsIWidget::Modifiers aModifierFlags,
-                                      nsIObserver* aObserver) override;
-
-  nsresult SynthesizeNativeMouseMove(LayoutDeviceIntPoint aPoint,
-                                     nsIObserver* aObserver) override {
+  nsresult SynthesizeNativeMouseMove(
+      LayoutDeviceIntPoint aPoint,
+      nsISynthesizedEventCallback* aCallback) override {
     return SynthesizeNativeMouseEvent(
         aPoint, NativeMouseMessage::Move, mozilla::MouseButton::eNotPressed,
-        nsIWidget::Modifiers::NO_MODIFIERS, aObserver);
+        nsIWidget::Modifiers::NO_MODIFIERS, aCallback);
   }
 
   nsresult SynthesizeNativeMouseScrollEvent(
       LayoutDeviceIntPoint aPoint, uint32_t aNativeMessage, double aDeltaX,
       double aDeltaY, double aDeltaZ, uint32_t aModifierFlags,
-      uint32_t aAdditionalFlags, nsIObserver* aObserver) override;
+      uint32_t aAdditionalFlags,
+      nsISynthesizedEventCallback* aCallback) override;
 
-  nsresult SynthesizeNativeTouchPoint(uint32_t aPointerId,
-                                      TouchPointerState aPointerState,
-                                      LayoutDeviceIntPoint aPoint,
-                                      double aPointerPressure,
-                                      uint32_t aPointerOrientation,
-                                      nsIObserver* aObserver) override;
+  nsresult SynthesizeNativeTouchPoint(
+      uint32_t aPointerId, TouchPointerState aPointerState,
+      LayoutDeviceIntPoint aPoint, double aPointerPressure,
+      uint32_t aPointerOrientation,
+      nsISynthesizedEventCallback* aCallback) override;
 
   nsresult SynthesizeNativeTouchPadPinch(TouchpadGesturePhase aEventPhase,
                                          float aScale,
                                          LayoutDeviceIntPoint aPoint,
                                          int32_t aModifierFlags) override;
 
-  nsresult SynthesizeNativeTouchpadPan(TouchpadGesturePhase aEventPhase,
-                                       LayoutDeviceIntPoint aPoint,
-                                       double aDeltaX, double aDeltaY,
-                                       int32_t aModifierFlags,
-                                       nsIObserver* aObserver) override;
+  nsresult SynthesizeNativeTouchpadPan(
+      TouchpadGesturePhase aEventPhase, LayoutDeviceIntPoint aPoint,
+      double aDeltaX, double aDeltaY, int32_t aModifierFlags,
+      nsISynthesizedEventCallback* aCallback) override;
 
   void GetCompositorWidgetInitData(
       mozilla::widget::CompositorWidgetInitData* aInitData) override;
 
-  nsresult SetNonClientMargins(const LayoutDeviceIntMargin&) override;
-  void SetDrawsInTitlebar(bool aState) override;
-  mozilla::LayoutDeviceIntCoord GetTitlebarRadius();
-  LayoutDeviceIntRect GetTitlebarRect();
+  void SetCustomTitlebar(bool) override;
   void UpdateWindowDraggingRegion(
       const LayoutDeviceIntRegion& aRegion) override;
 
+#ifdef MOZ_ENABLE_DBUS
+  void SetDBusMenuBar(RefPtr<mozilla::widget::DBusMenuBar> aDbusMenuBar);
+#endif
+
   // HiDPI scale conversion
   gint GdkCeiledScaleFactor();
-  bool UseFractionalScale() const;
   double FractionalScaleFactor();
 
   // To GDK
@@ -383,18 +423,24 @@ class nsWindow final : public nsBaseWidget {
   GdkPoint DevicePixelsToGdkPointRoundDown(const LayoutDeviceIntPoint&);
   GdkRectangle DevicePixelsToGdkSizeRoundUp(const LayoutDeviceIntSize&);
   GdkRectangle DevicePixelsToGdkRectRoundOut(const LayoutDeviceIntRect&);
+  GdkRectangle DevicePixelsToGdkRectRoundIn(const LayoutDeviceIntRect&);
 
   // From GDK
   int GdkCoordToDevicePixels(gint);
   LayoutDeviceIntPoint GdkPointToDevicePixels(const GdkPoint&);
   LayoutDeviceIntPoint GdkEventCoordsToDevicePixels(gdouble aX, gdouble aY);
   LayoutDeviceIntRect GdkRectToDevicePixels(const GdkRectangle&);
+  LayoutDeviceIntMargin GtkBorderToDevicePixels(const GtkBorder&);
 
   bool WidgetTypeSupportsAcceleration() override;
+  bool WidgetTypeSupportsNativeCompositing() override;
 
   nsresult SetSystemFont(const nsCString& aFontName) override;
   nsresult GetSystemFont(nsCString& aFontName) override;
 
+  void MaybeCreatePipResources();
+  void ClearPipResources();
+
   typedef enum {
     GTK_DECORATION_SYSTEM,  // CSD including shadows
     GTK_DECORATION_CLIENT,  // CSD without shadows
@@ -406,22 +452,23 @@ class nsWindow final : public nsBaseWidget {
    */
   static GtkWindowDecoration GetSystemGtkWindowDecoration();
 
-  static bool GetTopLevelWindowActiveState(nsIFrame* aFrame);
-  static bool TitlebarUseShapeMask();
-  bool IsRemoteContent() { return HasRemoteContent(); }
+  bool IsRemoteContent() const { return HasRemoteContent(); }
   void NativeMoveResizeWaylandPopupCallback(const GdkRectangle* aFinalSize,
                                             bool aFlippedX, bool aFlippedY);
   static bool IsToplevelWindowTransparent();
 
   static nsWindow* GetFocusedWindow();
 
+  mozilla::UniquePtr<mozilla::widget::WaylandSurfaceLock> LockSurface();
+
+  bool WaylandPipEnabled() const;
+
 #ifdef MOZ_WAYLAND
   // Use xdg-activation protocol to transfer focus from gFocusWindow to aWindow.
   static void TransferFocusToWaylandWindow(nsWindow* aWindow);
   void FocusWaylandWindow(const char* aTokenID);
 
-  bool GetCSDDecorationOffset(int* aDx, int* aDy);
-  void SetEGLNativeWindowSize(const LayoutDeviceIntSize& aEGLWindowSize);
+  bool SetEGLNativeWindowSize(const LayoutDeviceIntSize& aEGLWindowSize);
   void WaylandDragWorkaround(GdkEventButton* aEvent);
 
   void CreateCompositorVsyncDispatcher() override;
@@ -437,20 +484,7 @@ class nsWindow final : public nsBaseWidget {
   };
 #endif
 
-  typedef enum {
-    // WebRender compositor is enabled
-    COMPOSITOR_ENABLED,
-    // WebRender compositor is paused as we're repainting whole window and
-    // we're waiting for content process to update page content.
-    COMPOSITOR_PAUSED_FLICKERING
-  } WindowCompositorState;
-
-  // Pause compositor to avoid rendering artifacts from content process.
   void ResumeCompositorImpl();
-  void ResumeCompositorFlickering();
-  void ResumeCompositorFromCompositorThread();
-  void PauseCompositorFlickering();
-  bool IsWaitingForCompositorResume();
 
   // Force hide this window, remove compositor etc. to avoid
   // rendering queue blocking (see Bug 1782948).
@@ -462,24 +496,27 @@ class nsWindow final : public nsBaseWidget {
 
   static nsWindow* GetWindow(GdkWindow* window);
 
+  /**
+   * Dispatch accessible window activate event for the top level window
+   * accessible.
+   */
+  void DispatchActivateEventAccessible();
+
+  void GtkWidgetDestroyHandler(GtkWidget* aWidget);
+
+  void SetDragPopupSurface(RefPtr<gfxImageSurface> aDragPopupSurface,
+                           const LayoutDeviceIntRegion& aInvalidRegion);
+
  protected:
   virtual ~nsWindow();
 
   // event handling code
   void DispatchActivateEvent(void);
   void DispatchDeactivateEvent(void);
-  void MaybeDispatchResized();
   void DispatchPanGesture(mozilla::PanGestureInput& aPanInput);
 
   void RegisterTouchWindow() override;
 
-  nsCOMPtr<nsIWidget> mParent;
-  PopupType mPopupHint{};
-  int mWindowScaleFactor = 1;
-
-  void UpdateAlpha(mozilla::gfx::SourceSurface* aSourceSurface,
-                   nsIntRect aBoundsRect);
-
   void NativeMoveResize(bool aMoved, bool aResized);
 
   void NativeShow(bool aAction);
@@ -492,20 +529,23 @@ class nsWindow final : public nsBaseWidget {
 
   void TryToShowNativeWindowMenu(GdkEventButton* aEvent);
 
+  bool DoTitlebarAction(mozilla::LookAndFeel::TitlebarEvent aEvent,
+                        GdkEventButton* aButtonEvent);
+
   void WaylandStartVsync();
   void WaylandStopVsync();
   void DestroyChildWindows();
   GtkWidget* GetToplevelWidget() const;
   nsWindow* GetContainerWindow() const;
   Window GetX11Window();
-  bool GetShapedState();
   void EnsureGdkWindow();
   void SetUrgencyHint(GtkWidget* top_window, bool state);
   void SetDefaultIcon(void);
   void SetWindowDecoration(BorderStyle aStyle);
   void InitButtonEvent(mozilla::WidgetMouseEvent& aEvent,
                        GdkEventButton* aGdkEvent,
-                       const mozilla::LayoutDeviceIntPoint& aRefPoint);
+                       const mozilla::LayoutDeviceIntPoint& aRefPoint,
+                       bool isEraser = false);
   bool CheckForRollup(gdouble aMouseX, gdouble aMouseY, bool aIsWheel,
                       bool aAlwaysRollup);
   void RollupAllMenus() { CheckForRollup(0, 0, false, true); }
@@ -530,7 +570,7 @@ class nsWindow final : public nsBaseWidget {
   GtkTextDirection GetTextDirection();
 
   bool DrawsToCSDTitlebar() const;
-  void AddCSDDecorationSize(int* aWidth, int* aHeight);
+  bool ToplevelUsesCSD() const;
 
   void CreateAndPutGdkScrollEvent(mozilla::LayoutDeviceIntPoint aPoint,
                                   double aDeltaX, double aDeltaY);
@@ -543,12 +583,13 @@ class nsWindow final : public nsBaseWidget {
   GtkWidget* mShell = nullptr;
   MozContainer* mContainer = nullptr;
   GdkWindow* mGdkWindow = nullptr;
+#ifdef MOZ_WAYLAND
+  RefPtr<mozilla::widget::WaylandSurface> mSurface;
+#endif
+  RefPtr<gfxImageSurface> mDragPopupSurface;
+  LayoutDeviceIntRegion mDragPopupSurfaceRegion;
+
   PlatformCompositorWidgetDelegate* mCompositorWidgetDelegate = nullptr;
-  mozilla::Atomic<WindowCompositorState, mozilla::Relaxed> mCompositorState{
-      COMPOSITOR_ENABLED};
-  // This is used in COMPOSITOR_PAUSED_FLICKERING mode only to resume compositor
-  // in some reasonable time when page content is not updated.
-  guint mCompositorPauseTimeoutID = 0;
 
   // The actual size mode that's in effect.
   nsSizeMode mSizeMode = nsSizeMode_Normal;
@@ -559,6 +600,12 @@ class nsWindow final : public nsBaseWidget {
 
   float mAspectRatio = 0.0f;
   float mAspectRatioSaved = 0.0f;
+  mozilla::Maybe<GtkOrientation> mAspectResizer;
+  GdkPoint mLastResizePoint{0, 0};
+
+  // Keep in sync with WaylandSurface::sNoScale
+  constexpr static const int sNoScale = -1;
+  int mCeiledScaleFactor = sNoScale;
 
   // The size requested, which might not be reflected in mBounds.  Used in
   // WaylandPopupSetDirectPosition() to remember intended size for popup
@@ -566,9 +613,12 @@ class nsWindow final : public nsBaseWidget {
   // to remember a size requested while waiting for moved-to-rect when
   // OnSizeAllocate() might change mBounds.Size().
   LayoutDeviceIntSize mLastSizeRequest;
-  LayoutDeviceIntPoint mClientOffset;
-  // Indicates a new size that still needs to be dispatched.
-  LayoutDeviceIntSize mNeedsDispatchSize = LayoutDeviceIntSize(-1, -1);
+  // Same but for positioning. Used to track move requests.
+  LayoutDeviceIntPoint mLastMoveRequest;
+  // Margin from outer bounds to inner bounds _including CSD decorations_.
+  LayoutDeviceIntMargin mClientMargin;
+  // The part of mClientMargin that comes from our CSD decorations.
+  LayoutDeviceIntMargin mCsdMargin;
 
   // This field omits duplicate scroll events caused by GNOME bug 726878.
   guint32 mLastScrollEventTime = GDK_CURRENT_TIME;
@@ -591,7 +641,7 @@ class nsWindow final : public nsBaseWidget {
   // Used for synthesizing touchpad pan gestures
   struct TouchpadPanGestureState {
     mozilla::Maybe<TouchpadGesturePhase> mTouchpadGesturePhase;
-    uint64_t mSavedObserver = 0;
+    mozilla::Maybe<uint64_t> mSavedCallbackId;
   };
 
   // Used for synthesizing touchpad pan gestures
@@ -616,17 +666,16 @@ class nsWindow final : public nsBaseWidget {
   static GdkCursor* gsGtkCursorCache[eCursorCount];
 
   // If true, draw our own window titlebar.
-  //
-  // Needs to be atomic because GetTitlebarRect() gets called from non-main
-  // threads.
-  //
-  // FIXME(emilio): GetTitlebarRect() reads other things that TSAN doesn't
-  // catch because mDrawInTitlebar is false on automation ~always. We should
-  // probably make GetTitlebarRect() simpler / properly thread-safe.
-  mozilla::Atomic<bool, mozilla::Relaxed> mDrawInTitlebar{false};
+  bool mDrawInTitlebar = false;
 
+  // This mutex protect window visibility changes.
+  mozilla::Mutex mWindowVisibilityMutex;
+
+  // This track real window visibility from OS perspective.
+  // It's set by OnMap/OnUnmap which is based on Gtk events.
+  mozilla::Atomic<bool, mozilla::Relaxed> mIsMapped;
   // Has this widget been destroyed yet?
-  bool mIsDestroyed;
+  mozilla::Atomic<bool, mozilla::Relaxed> mIsDestroyed;
   // mIsShown tracks requested visible status from browser perspective, i.e.
   // if the window should be visible or now.
   bool mIsShown : 1;
@@ -636,9 +685,6 @@ class nsWindow final : public nsBaseWidget {
   // that the window is not actually visible but we report to browser that
   // it is visible (mIsShown == true).
   bool mNeedsShow : 1;
-  // This track real window visibility from OS perspective.
-  // It's set by OnMap/OnUnrealize which is based on Gtk events.
-  bool mIsMapped : 1;
   // is this widget enabled?
   bool mEnabled : 1;
   // has the native window for this been created yet?
@@ -647,33 +693,27 @@ class nsWindow final : public nsBaseWidget {
   bool mHandleTouchEvent : 1;
   // true if this is a drag and drop feedback popup
   bool mIsDragPopup : 1;
-  bool mWindowScaleFactorChanged : 1;
   bool mCompositedScreen : 1;
   bool mIsAccelerated : 1;
+  bool mIsAlert : 1;
   bool mWindowShouldStartDragging : 1;
   bool mHasMappedToplevel : 1;
-  bool mRetryPointerGrab : 1;
   bool mPanInProgress : 1;
-  // Use dedicated GdkWindow for mContainer
-  bool mDrawToContainer : 1;
+  bool mPendingBoundsChange : 1;
+  // Whether our pending bounds change event might change the window CSD margin.
+  // This is needed because we might get two configures (one for mShell, one
+  // for mContainer's window) in quick succession, which might cause us to send
+  // spurious sequences of resizes if we don't do this on some compositors
+  // (older mutter at least).
+  bool mPendingBoundsChangeMayChangeCsdMargin : 1;
   // Draw titlebar with :backdrop css state (inactive/unfocused).
   bool mTitlebarBackdropState : 1;
-  // It's PictureInPicture window.
-  bool mIsPIPWindow : 1;
-  // It's undecorated popup utility window, without resizers/titlebar,
-  // movable by mouse. Used on Wayland for popups without
-  // parent (for instance WebRTC sharing indicator, notifications).
-  bool mIsWaylandPanelWindow : 1;
-  // It's child window, i.e. window which is nested in parent window.
-  // This is obsoleted and should not be used.
-  // We use GdkWindow hierarchy for such windows.
-  bool mIsChildWindow : 1;
   bool mAlwaysOnTop : 1;
-  bool mNoAutoHide : 1;
   bool mIsTransparent : 1;
   // We can expect at least one size-allocate event after early resizes.
   bool mHasReceivedSizeAllocate : 1;
   bool mWidgetCursorLocked : 1;
+  bool mUndecorated : 1;
 
   /*  Gkt creates popup in two incarnations - wl_subsurface and xdg_popup.
    *  Kind of popup is choosen before GdkWindow is mapped so we can change
@@ -715,10 +755,6 @@ class nsWindow final : public nsBaseWidget {
    */
   bool mHiddenPopupPositioned : 1;
 
-  // The transparency bitmap is used instead of ARGB visual for toplevel
-  // window to draw titlebar.
-  bool mTransparencyBitmapForTitlebar : 1;
-
   // True when we're on compositing window manager and this
   // window is using visual with alpha channel.
   bool mHasAlphaVisual : 1;
@@ -781,13 +817,6 @@ class nsWindow final : public nsBaseWidget {
   // Whether we need to retry capturing the mouse because we' re not mapped yet.
   bool mNeedsToRetryCapturingMouse : 1;
 
-  // This bitmap tracks which pixels are transparent. We don't support
-  // full translucency at this time; each pixel is either fully opaque
-  // or fully transparent.
-  gchar* mTransparencyBitmap = nullptr;
-  int32_t mTransparencyBitmapWidth = 0;
-  int32_t mTransparencyBitmapHeight = 0;
-
   // all of our DND stuff
   void InitDragEvent(mozilla::WidgetDragEvent& aEvent);
 
@@ -795,15 +824,10 @@ class nsWindow final : public nsBaseWidget {
 
   InputRegion mInputRegion;
 
-  static bool DragInProgress(void);
+  bool DragInProgress(void);
 
   void DispatchMissedButtonReleases(GdkEventCrossing* aGdkEvent);
 
-  // When window widget gets mapped/unmapped we need to configure
-  // underlying GdkWindow properly. Otherwise we'll end up with
-  // rendering to released window.
-  void ConfigureGdkWindow();
-  void ReleaseGdkWindow();
   void ConfigureCompositor();
 
   bool IsAlwaysUndecoratedWindow() const;
@@ -830,7 +854,6 @@ class nsWindow final : public nsBaseWidget {
   GdkPoint WaylandGetParentPosition();
   bool WaylandPopupConfigure();
   bool WaylandPopupIsAnchored();
-  bool WaylandPopupIsMenu();
   bool WaylandPopupIsContextMenu();
   bool WaylandPopupIsPermanent();
   // First popup means it's attached directly to toplevel window
@@ -864,13 +887,13 @@ class nsWindow final : public nsBaseWidget {
                                      GdkPoint* aOffset);
   bool WaylandPopupAnchorAdjustForParentPopup(GdkRectangle* aPopupAnchor,
                                               GdkPoint* aOffset);
-  nsWindow* WaylandPopupGetTopmostWindow();
+  nsWindow* GetTopmostWindow();
   bool IsPopupInLayoutPopupChain(nsTArray<nsIWidget*>* aLayoutWidgetHierarchy,
                                  bool aMustMatchParent);
   void WaylandPopupMarkAsClosed();
   void WaylandPopupRemoveClosedPopups();
   void WaylandPopupSetDirectPosition();
-  bool WaylandPopupFitsToplevelWindow(bool aMove);
+  bool WaylandPopupFitsToplevelWindow();
   const WaylandPopupMoveToRectParams WaylandPopupGetPositionFromLayout();
   void WaylandPopupPropagateChangesToLayout(bool aMove, bool aResize);
   nsWindow* WaylandPopupFindLast(nsWindow* aPopup);
@@ -907,6 +930,27 @@ class nsWindow final : public nsBaseWidget {
   // in following popup operations unless mLayoutPopupSizeCleared is set.
   LayoutDeviceIntSize mMoveToRectPopupSize;
 
+#ifdef MOZ_ENABLE_DBUS
+  RefPtr<mozilla::widget::DBusMenuBar> mDBusMenuBar;
+#endif
+
+  struct LastMouseCoordinates {
+    template <typename Event>
+    void Set(Event* aEvent) {
+      mX = aEvent->x;
+      mY = aEvent->y;
+      mRootX = aEvent->x_root;
+      mRootY = aEvent->y_root;
+    }
+
+    float mX = 0.0f, mY = 0.0f;
+    float mRootX = 0.0f, mRootY = 0.0f;
+  } mLastMouseCoordinates;
+
+  // We don't want to fire scroll event with the same timestamp as
+  // smooth scroll event.
+  guint32 mLastSmoothScrollEventTime = GDK_CURRENT_TIME;
+
   /**
    * |mIMContext| takes all IME related stuff.
    *
@@ -944,12 +988,6 @@ class nsWindow final : public nsBaseWidget {
    */
   void DispatchEventToRootAccessible(uint32_t aEventType);
 
-  /**
-   * Dispatch accessible window activate event for the top level window
-   * accessible.
-   */
-  void DispatchActivateEventAccessible();
-
   /**
    * Dispatch accessible window deactivate event for the top level window
    * accessible.
@@ -977,10 +1015,20 @@ class nsWindow final : public nsBaseWidget {
 
   void SetUserTimeAndStartupTokenForActivatedWindow();
 
+  void KioskLockOnMonitor();
+
+  void EmulateResizeDrag(GdkEventMotion* aEvent);
+
+  void RequestRepaint(LayoutDeviceIntRegion& aRepaintRegion);
+
+  bool DrawDragPopupSurface(cairo_t* cr);
+
 #ifdef MOZ_X11
-  typedef enum {GTK_WIDGET_COMPOSIDED_DEFAULT = 0,
-                GTK_WIDGET_COMPOSIDED_DISABLED = 1,
-                GTK_WIDGET_COMPOSIDED_ENABLED = 2} WindowComposeRequest;
+  typedef enum {
+    GTK_WIDGET_COMPOSITED_DEFAULT = 0,
+    GTK_WIDGET_COMPOSITED_DISABLED = 1,
+    GTK_WIDGET_COMPOSITED_ENABLED = 2
+  } WindowComposeRequest;
   void SetCompositorHint(WindowComposeRequest aState);
   bool ConfigureX11GLVisual();
 #endif
@@ -996,9 +1044,18 @@ class nsWindow final : public nsBaseWidget {
   nsCString mWindowActivationTokenFromEnv;
   mozilla::widget::WindowSurfaceProvider mSurfaceProvider;
   GdkDragContext* mSourceDragContext = nullptr;
-#if MOZ_LOGGING
-  LayoutDeviceIntRect mLastLoggedBoundSize;
-  int mLastLoggedScale = -1;
+  mozilla::Sides mResizableEdges{mozilla::SideBits::eAll};
+  // Running in kiosk mode and requested to stay on specified monitor.
+  // If monitor is removed minimize the window.
+  mozilla::Maybe<int> mKioskMonitor;
+  LayoutDeviceIntRegion mOpaqueRegion MOZ_GUARDED_BY(mOpaqueRegionLock);
+  mutable mozilla::RWLock mOpaqueRegionLock{"nsWindow::mOpaqueRegion"};
+#ifdef MOZ_WAYLAND
+  struct {
+    struct xdg_surface* mXdgSurface = nullptr;
+    struct xx_pip_v1* mPipSurface = nullptr;
+    LayoutDeviceIntSize mConfigureSize;
+  } mPipResources;
 #endif
 };
 
