@@ -8,7 +8,6 @@
 #include "mozilla/dom/Document.h"
 #include "nsIContent.h"
 #include "nsIFrame.h"
-#include "nsIScrollableFrame.h"
 #include "nsLayoutUtils.h"
 #include "nsNumberControlFrame.h"
 #include "nsPresContext.h"
@@ -17,18 +16,16 @@
 #include "nsStyleConsts.h"
 #include "nsPIDOMWindow.h"
 #include "nsProgressFrame.h"
-#include "nsMeterFrame.h"
-#include "nsMenuFrame.h"
 #include "nsRangeFrame.h"
 #include "nsCSSRendering.h"
 #include "ImageContainer.h"
 #include "mozilla/ComputedStyle.h"
-#include "mozilla/EventStates.h"
 #include "mozilla/dom/Element.h"
 #include "mozilla/dom/HTMLBodyElement.h"
 #include "mozilla/dom/HTMLInputElement.h"
 #include "mozilla/dom/HTMLProgressElement.h"
 #include "mozilla/PresShell.h"
+#include "mozilla/ScrollContainerFrame.h"
 #include "mozilla/StaticPrefs_layout.h"
 #include "mozilla/dom/DocumentInlines.h"
 #include <algorithm>
@@ -40,44 +37,34 @@ nsNativeTheme::nsNativeTheme() : mAnimatedContentTimeout(UINT32_MAX) {}
 
 NS_IMPL_ISUPPORTS(nsNativeTheme, nsITimerCallback, nsINamed)
 
-/* static */ EventStates nsNativeTheme::GetContentState(
+/* static */ ElementState nsNativeTheme::GetContentState(
     nsIFrame* aFrame, StyleAppearance aAppearance) {
   if (!aFrame) {
-    return EventStates();
+    return ElementState();
   }
 
   nsIContent* frameContent = aFrame->GetContent();
   if (!frameContent || !frameContent->IsElement()) {
-    return EventStates();
+    return ElementState();
   }
 
   const bool isXULElement = frameContent->IsXULElement();
   if (isXULElement) {
-    if (aAppearance == StyleAppearance::CheckboxLabel ||
-        aAppearance == StyleAppearance::RadioLabel) {
-      aFrame = aFrame->GetParent()->GetParent();
-      frameContent = aFrame->GetContent();
-    } else if (aAppearance == StyleAppearance::Checkbox ||
-               aAppearance == StyleAppearance::Radio ||
-               aAppearance == StyleAppearance::ToolbarbuttonDropdown ||
-               aAppearance == StyleAppearance::Treeheadersortarrow ||
-               aAppearance == StyleAppearance::ButtonArrowPrevious ||
-               aAppearance == StyleAppearance::ButtonArrowNext ||
-               aAppearance == StyleAppearance::ButtonArrowUp ||
-               aAppearance == StyleAppearance::ButtonArrowDown) {
+    if (aAppearance == StyleAppearance::Checkbox ||
+        aAppearance == StyleAppearance::Radio) {
       aFrame = aFrame->GetParent();
       frameContent = aFrame->GetContent();
     }
     MOZ_ASSERT(frameContent && frameContent->IsElement());
   }
 
-  EventStates flags = frameContent->AsElement()->StyleState();
+  ElementState flags = frameContent->AsElement()->StyleState();
   nsNumberControlFrame* numberControlFrame =
       nsNumberControlFrame::GetNumberControlFrameForSpinButton(aFrame);
   if (numberControlFrame &&
       numberControlFrame->GetContent()->AsElement()->StyleState().HasState(
-          NS_EVENT_STATE_DISABLED)) {
-    flags |= NS_EVENT_STATE_DISABLED;
+          ElementState::DISABLED)) {
+    flags |= ElementState::DISABLED;
   }
 
   if (!isXULElement) {
@@ -85,42 +72,45 @@ NS_IMPL_ISUPPORTS(nsNativeTheme, nsITimerCallback, nsINamed)
   }
 
   if (CheckBooleanAttr(aFrame, nsGkAtoms::disabled)) {
-    flags |= NS_EVENT_STATE_DISABLED;
+    flags |= ElementState::DISABLED;
   }
 
   switch (aAppearance) {
-    case StyleAppearance::RadioLabel:
     case StyleAppearance::Radio: {
       if (CheckBooleanAttr(aFrame, nsGkAtoms::focused)) {
-        flags |= NS_EVENT_STATE_FOCUS;
+        flags |= ElementState::FOCUS;
         nsPIDOMWindowOuter* window =
             aFrame->GetContent()->OwnerDoc()->GetWindow();
         if (window && window->ShouldShowFocusRing()) {
-          flags |= NS_EVENT_STATE_FOCUSRING;
+          flags |= ElementState::FOCUSRING;
         }
       }
-      if (CheckBooleanAttr(aFrame, nsGkAtoms::selected)) {
-        flags |= NS_EVENT_STATE_CHECKED;
+      if (CheckBooleanAttr(aFrame, nsGkAtoms::selected) ||
+          CheckBooleanAttr(aFrame, nsGkAtoms::checked)) {
+        flags |= ElementState::CHECKED;
       }
       break;
     }
-    case StyleAppearance::CheckboxLabel:
     case StyleAppearance::Checkbox: {
       if (CheckBooleanAttr(aFrame, nsGkAtoms::checked)) {
-        flags |= NS_EVENT_STATE_CHECKED;
+        flags |= ElementState::CHECKED;
       } else if (CheckBooleanAttr(aFrame, nsGkAtoms::indeterminate)) {
-        flags |= NS_EVENT_STATE_INDETERMINATE;
+        flags |= ElementState::INDETERMINATE;
       }
       break;
     }
-    case StyleAppearance::MenulistButton:
+    case StyleAppearance::Toolbarbutton:
+      if (CheckBooleanAttr(aFrame, nsGkAtoms::open)) {
+        flags |= ElementState::HOVER | ElementState::ACTIVE;
+      }
+      break;
     case StyleAppearance::Menulist:
     case StyleAppearance::NumberInput:
     case StyleAppearance::Textfield:
-    case StyleAppearance::Searchfield:
+    case StyleAppearance::PasswordInput:
     case StyleAppearance::Textarea: {
       if (CheckBooleanAttr(aFrame, nsGkAtoms::focused)) {
-        flags |= NS_EVENT_STATE_FOCUS | NS_EVENT_STATE_FOCUSRING;
+        flags |= ElementState::FOCUS | ElementState::FOCUSRING;
       }
       break;
     }
@@ -138,8 +128,7 @@ bool nsNativeTheme::CheckBooleanAttr(nsIFrame* aFrame, nsAtom* aAtom) {
   nsIContent* content = aFrame->GetContent();
   if (!content || !content->IsElement()) return false;
 
-  if (content->IsHTMLElement())
-    return content->AsElement()->HasAttr(kNameSpaceID_None, aAtom);
+  if (content->IsHTMLElement()) return content->AsElement()->HasAttr(aAtom);
 
   // For XML/XUL elements, an attribute must be equal to the literal
   // string "true" to be counted as true.  An empty string should _not_
@@ -157,7 +146,7 @@ int32_t nsNativeTheme::CheckIntAttr(nsIFrame* aFrame, nsAtom* aAtom,
   if (!content || !content->IsElement()) return defaultValue;
 
   nsAutoString attr;
-  content->AsElement()->GetAttr(kNameSpaceID_None, aAtom, attr);
+  content->AsElement()->GetAttr(aAtom, attr);
   nsresult err;
   int32_t value = attr.ToInteger(&err);
   if (attr.IsEmpty() || NS_FAILED(err)) return defaultValue;
@@ -193,12 +182,13 @@ bool nsNativeTheme::IsButtonTypeMenu(nsIFrame* aFrame) {
 }
 
 bool nsNativeTheme::IsPressedButton(nsIFrame* aFrame) {
-  EventStates eventState =
-      GetContentState(aFrame, StyleAppearance::Toolbarbutton);
-  if (eventState.HasState(NS_EVENT_STATE_DISABLED)) return false;
+  ElementState state = GetContentState(aFrame, StyleAppearance::Toolbarbutton);
+  if (state.HasState(ElementState::DISABLED)) {
+    return false;
+  }
 
   return IsOpenButton(aFrame) ||
-         eventState.HasAllStates(NS_EVENT_STATE_ACTIVE | NS_EVENT_STATE_HOVER);
+         state.HasAllStates(ElementState::ACTIVE | ElementState::HOVER);
 }
 
 bool nsNativeTheme::IsWidgetStyled(nsPresContext* aPresContext,
@@ -209,56 +199,14 @@ bool nsNativeTheme::IsWidgetStyled(nsPresContext* aPresContext,
     return false;
   }
 
-  // Resizers have some special handling, dependent on whether in a scrollable
-  // container or not. If so, use the scrollable container's to determine
-  // whether the style is overriden instead of the resizer. This allows a
-  // non-native transparent resizer to be used instead. Otherwise, we just
-  // fall through and return false.
-  if (aAppearance == StyleAppearance::Resizer) {
-    nsIFrame* parentFrame = aFrame->GetParent();
-    if (parentFrame && parentFrame->IsScrollFrame()) {
-      // if the parent is a scrollframe, the resizer should be native themed
-      // only if the scrollable area doesn't override the widget style.
-      //
-      // note that the condition below looks a bit suspect but it's the right
-      // one. If there's no valid appearance, then we should return true, it's
-      // effectively the same as if it had overridden the appearance.
-      parentFrame = parentFrame->GetParent();
-      if (!parentFrame) {
-        return false;
-      }
-      auto parentAppearance =
-          parentFrame->StyleDisplay()->EffectiveAppearance();
-      return parentAppearance == StyleAppearance::None ||
-             IsWidgetStyled(aPresContext, parentFrame, parentAppearance);
-    }
-  }
-
   /**
    * Progress bar appearance should be the same for the bar and the container
-   * frame. nsProgressFrame owns the logic and will tell us what we should do.
+   * frame.
    */
-  if (aAppearance == StyleAppearance::Progresschunk ||
-      aAppearance == StyleAppearance::ProgressBar) {
-    nsProgressFrame* progressFrame = do_QueryFrame(
-        aAppearance == StyleAppearance::Progresschunk ? aFrame->GetParent()
-                                                      : aFrame);
-    if (progressFrame) {
-      return !progressFrame->ShouldUseNativeStyle();
-    }
-  }
-
-  /**
-   * Meter bar appearance should be the same for the bar and the container
-   * frame. nsMeterFrame owns the logic and will tell us what we should do.
-   */
-  if (aAppearance == StyleAppearance::Meterchunk ||
+  if (aAppearance == StyleAppearance::ProgressBar ||
       aAppearance == StyleAppearance::Meter) {
-    nsMeterFrame* meterFrame = do_QueryFrame(
-        aAppearance == StyleAppearance::Meterchunk ? aFrame->GetParent()
-                                                   : aFrame);
-    if (meterFrame) {
-      return !meterFrame->ShouldUseNativeStyle();
+    if (nsProgressFrame* progressFrame = do_QueryFrame(aFrame)) {
+      return !progressFrame->ShouldUseNativeStyle();
     }
   }
 
@@ -267,12 +215,8 @@ bool nsNativeTheme::IsWidgetStyled(nsPresContext* aPresContext,
    * comes to native theming (either all parts, or no parts, are themed).
    * nsRangeFrame owns the logic and will tell us what we should do.
    */
-  if (aAppearance == StyleAppearance::Range ||
-      aAppearance == StyleAppearance::RangeThumb) {
-    nsRangeFrame* rangeFrame = do_QueryFrame(
-        aAppearance == StyleAppearance::RangeThumb ? aFrame->GetParent()
-                                                   : aFrame);
-    if (rangeFrame) {
+  if (aAppearance == StyleAppearance::Range) {
+    if (nsRangeFrame* rangeFrame = do_QueryFrame(aFrame)) {
       return !rangeFrame->ShouldUseNativeStyle();
     }
   }
@@ -323,10 +267,14 @@ nsNativeTheme::TreeSortDirection nsNativeTheme::GetTreeSortDirection(
 }
 
 bool nsNativeTheme::IsLastTreeHeaderCell(nsIFrame* aFrame) {
-  if (!aFrame) return false;
+  if (!aFrame) {
+    return false;
+  }
 
-  // A tree column picker is always the last header cell.
-  if (aFrame->GetContent()->IsXULElement(nsGkAtoms::treecolpicker)) return true;
+  // A tree column picker button is always the last header cell.
+  if (aFrame->GetContent()->IsXULElement(nsGkAtoms::button)) {
+    return true;
+  }
 
   // Find the parent tree.
   nsIContent* parent = aFrame->GetContent()->GetParent();
@@ -352,11 +300,10 @@ bool nsNativeTheme::IsBottomTab(nsIFrame* aFrame) {
 
   nsAutoString classStr;
   if (aFrame->GetContent()->IsElement()) {
-    aFrame->GetContent()->AsElement()->GetAttr(kNameSpaceID_None,
-                                               nsGkAtoms::_class, classStr);
+    aFrame->GetContent()->AsElement()->GetAttr(nsGkAtoms::_class, classStr);
   }
   // FIXME: This looks bogus, shouldn't this be looking at GetClasses()?
-  return !classStr.IsEmpty() && classStr.Find("tab-bottom") != kNotFound;
+  return !classStr.IsEmpty() && classStr.Find(u"tab-bottom") != kNotFound;
 }
 
 bool nsNativeTheme::IsFirstTab(nsIFrame* aFrame) {
@@ -469,8 +416,7 @@ bool nsNativeTheme::QueueAnimatedContentForRefresh(nsIContent* aContent,
     }
 
     if (XRE_IsContentProcess() && NS_IsMainThread()) {
-      mAnimatedContentTimer->SetTarget(
-          aContent->OwnerDoc()->EventTargetFor(TaskCategory::Other));
+      mAnimatedContentTimer->SetTarget(GetMainThreadSerialEventTarget());
     }
     rv = mAnimatedContentTimer->InitWithCallback(this, timeout,
                                                  nsITimer::TYPE_ONE_SHOT);
@@ -547,55 +493,33 @@ bool nsNativeTheme::IsRangeHorizontal(nsIFrame* aFrame) {
   return aFrame->GetSize().width >= aFrame->GetSize().height;
 }
 
-static nsIFrame* GetBodyFrame(nsIFrame* aCanvasFrame) {
-  nsIContent* body = aCanvasFrame->PresContext()->Document()->GetBodyElement();
-  if (!body) {
-    return nullptr;
-  }
-  return body->GetPrimaryFrame();
-}
-
 /* static */
-bool nsNativeTheme::IsDarkBackground(nsIFrame* aFrame) {
+bool nsNativeTheme::IsDarkBackgroundForScrollbar(nsIFrame* aFrame) {
   // Try to find the scrolled frame. Note that for stuff like xul <tree> there
   // might be none.
   {
     nsIFrame* frame = aFrame;
-    nsIScrollableFrame* scrollFrame = nullptr;
-    while (!scrollFrame && frame) {
-      scrollFrame = frame->GetScrollTargetFrame();
+    ScrollContainerFrame* scrollContainerFrame = nullptr;
+    while (!scrollContainerFrame && frame) {
+      scrollContainerFrame = frame->GetScrollTargetFrame();
       frame = frame->GetParent();
     }
-    if (scrollFrame) {
-      aFrame = scrollFrame->GetScrolledFrame();
+    if (scrollContainerFrame) {
+      aFrame = scrollContainerFrame->GetScrolledFrame();
     } else {
       // Leave aFrame untouched.
     }
   }
 
-  auto backgroundFrame = nsCSSRendering::FindNonTransparentBackgroundFrame(
-      aFrame, /* aStopAtThemed = */ false);
-  if (!backgroundFrame.mFrame) {
-    return false;
-  }
-
-  nscolor color = backgroundFrame.mFrame->StyleBackground()->BackgroundColor(
-      backgroundFrame.mFrame);
-
-  if (backgroundFrame.mIsForCanvas) {
-    // For canvas frames, prefer to look at the body first, because the body
-    // background color is most likely what will be visible as the background
-    // color of the page, even if the html element has a different background
-    // color which prevents that of the body frame to propagate to the viewport.
-    if (nsIFrame* bodyFrame = GetBodyFrame(aFrame)) {
-      nscolor bodyColor =
-          bodyFrame->StyleBackground()->BackgroundColor(bodyFrame);
-      if (NS_GET_A(bodyColor)) {
-        color = bodyColor;
-      }
-    }
-  }
+  return IsDarkBackground(aFrame);
+}
 
+/* static */
+bool nsNativeTheme::IsDarkBackground(nsIFrame* aFrame) {
+  auto color =
+      nsCSSRendering::FindEffectiveBackgroundColor(
+          aFrame, /* aStopAtThemed = */ false, /* aPreferBodyToCanvas = */ true)
+          .mColor;
   return LookAndFeel::IsDarkColor(color);
 }
 
@@ -610,11 +534,24 @@ bool nsNativeTheme::IsWidgetScrollbarPart(StyleAppearance aAppearance) {
     case StyleAppearance::ScrollbarbuttonRight:
     case StyleAppearance::ScrollbarthumbVertical:
     case StyleAppearance::ScrollbarthumbHorizontal:
-    case StyleAppearance::ScrollbartrackHorizontal:
-    case StyleAppearance::ScrollbartrackVertical:
     case StyleAppearance::Scrollcorner:
       return true;
     default:
       return false;
   }
 }
+
+/*static*/
+bool nsNativeTheme::IsWidgetAlwaysNonNative(nsIFrame* aFrame,
+                                            StyleAppearance aAppearance) {
+  return IsWidgetScrollbarPart(aAppearance) ||
+         aAppearance == StyleAppearance::FocusOutline ||
+         aAppearance == StyleAppearance::SpinnerUpbutton ||
+         aAppearance == StyleAppearance::SpinnerDownbutton ||
+         aAppearance == StyleAppearance::Toolbarbutton ||
+         aAppearance == StyleAppearance::ProgressBar ||
+         aAppearance == StyleAppearance::Meter ||
+         aAppearance == StyleAppearance::Range ||
+         aAppearance == StyleAppearance::Listbox ||
+         (aFrame && aFrame->StyleUI()->mMozTheme == StyleMozTheme::NonNative);
+}
