@@ -18,7 +18,6 @@
 #include "mozilla/MiscEvents.h"
 #include "mozilla/TextEvents.h"
 #include "mozilla/ToString.h"
-#include "mozilla/WindowsVersion.h"
 
 #ifndef IME_PROP_ACCEPT_WIDE_VKEY
 #  define IME_PROP_ACCEPT_WIDE_VKEY 0x20
@@ -47,8 +46,6 @@
 // Therefore you shouldn't use `LogLevel::Verbose` for logging usual behavior.
 extern mozilla::LazyLogModule gIMELog;
 
-static const char* GetBoolName(bool aBool) { return aBool ? "true" : "false"; }
-
 static void HandleSeparator(nsACString& aDesc) {
   if (!aDesc.IsEmpty()) {
     aDesc.AppendLiteral(" | ");
@@ -181,8 +178,8 @@ void IMEContext::Clear() {
 
 static UINT sWM_MSIME_MOUSE = 0;  // mouse message for MSIME 98/2000
 
-WritingMode IMMHandler::sWritingModeOfCompositionFont;
-nsString IMMHandler::sIMEName;
+MOZ_RUNINIT WritingMode IMMHandler::sWritingModeOfCompositionFont;
+MOZ_RUNINIT nsString IMMHandler::sIMEName;
 UINT IMMHandler::sCodePage = 0;
 DWORD IMMHandler::sIMEProperty = 0;
 DWORD IMMHandler::sIMEUIProperty = 0;
@@ -224,8 +221,7 @@ bool IMMHandler::IsActiveIMEInBlockList() {
   // ATOK 2008 and ATOK 2009 crash when user just opens their open state.
   // ATOK 2010 isn't installable newly on Win 7 or later, but we have a lot of
   // crash reports.
-  if (IsWin8OrLater() &&
-      (IsATOK2006Active() || IsATOK2007Active() || IsATOK2008Active() ||
+  if ((IsATOK2006Active() || IsATOK2007Active() || IsATOK2008Active() ||
        IsATOK2009Active() || IsATOK2010Active())) {
     return true;
   }
@@ -357,7 +353,7 @@ IMENotificationRequests IMMHandler::GetIMENotificationRequests() {
 // used for checking the lParam of WM_IME_COMPOSITION
 #define IS_COMPOSING_LPARAM(lParam) \
   ((lParam) & (GCS_COMPSTR | GCS_COMPATTR | GCS_COMPCLAUSE | GCS_CURSORPOS))
-#define IS_COMMITTING_LPARAM(lParam) ((lParam)&GCS_RESULTSTR)
+#define IS_COMMITTING_LPARAM(lParam) ((lParam) & GCS_RESULTSTR)
 // Some IMEs (e.g., the standard IME for Korean) don't have caret position,
 // then, we should not set caret position to compositionchange event.
 #define NO_IME_CARET -1
@@ -1479,8 +1475,11 @@ bool IMMHandler::HandleDocumentFeed(nsWindow* aWindow, LPARAM lParam,
 
   // Get the focused paragraph, we decide that it starts from the previous CRLF
   // (or start of the editor) to the next one (or the end of the editor).
-  int32_t paragraphStart = str.RFind("\n", false, targetOffset, -1) + 1;
-  int32_t paragraphEnd = str.Find("\r", false, targetOffset + targetLength, -1);
+  int32_t paragraphStart = 0;
+  if (targetOffset > 0) {
+    paragraphStart = Substring(str, 0, targetOffset).RFind(u"\n") + 1;
+  }
+  int32_t paragraphEnd = str.Find(u"\r", targetOffset + targetLength);
   if (paragraphEnd < 0) {
     paragraphEnd = str.Length();
   }
@@ -2319,7 +2318,7 @@ nsresult IMMHandler::OnMouseButtonEvent(
   MOZ_LOG(gIMELog, LogLevel::Info,
           ("IMMHandler::OnMouseButtonEvent, x,y=%d,%d, offset=%d, "
            "positioning=%d",
-           cursorPos.x, cursorPos.y, offset, positioning));
+           cursorPos.x.value, cursorPos.y.value, offset, positioning));
 
   // send MS_MSIME_MOUSE message to default IME window.
   HWND imeWnd = ::ImmGetDefaultIMEWnd(aWindow->GetWindowHandle());
